// *********************************************************************************************
//
//  brtag.c - Broadcom EJTAG Debrick Utility v1.6r - by hugebird. 12/12/2008
//
// Default compile for Windows
//  Change...
//
// New for v1.6r 
//              -Added new 16MB Spansion S29GL-P Flash Chip to the list.      
//              -Added new 4MB/8MB/16MB Spansion S29GL-M Flash Chip to the list.
//                     - support show PPB sector protection status,  -probeonly /showppb
//                     - support clear PPB sector protection,  -probeonly /clearppb
//                     - allow erase whole flash with erase chip command,  -probeonly /erasechip
//              -Mass codes cleanup
//                     - allow load CPU configuration code to better support 16MB flash to BCM6358, 
//                       easily add other CPU initial codes in further.  /initcpu
//                     - add BCM6368 CPU ID to the list
//                     - update Pracc code, make R/W faster
//                     - erase window not align with sector boundary is better handled. /forcealign
//                     - allow custom set flash probing base address.     -probeonly /window:xxxxxxxx
//                     - adjust AMD flash command per latest S29GL128-P datasheet
//                     - file length and operation window routine for BCM63x8 updated.
//
//
// New for v1.1 -Added new 16MB Spansion Flash Chip Parts to the list:
//        Please note: 16MB flash is dived into 2x8MHz partitions,only first 8MB can be supported.
//                     - Spansion S29GL128N10TFI01         (16MB)

// Re-baseline the tornado version utility to my version 1.0 based on Tornado MOD
// New for BrJtag v1.0
//               Added Endianess selection for Router and ADSL modem
//                       -  useage: /revendian 
//               Added full support for ADSL modem chip
//                       - Broadcom 6345             
//                       - Broadcom 6338             
//                       - Broadcom 6348
//                       - Broadcom 6358
//               Added custom parallel port support
//               Added 64KB tiny CFE support
//                       - useage:  xxx:cfe /tinycfe     or   use xxx:tfe
//                       - useage:  xxx:kernel /tinycfe  or   use xxx:ternel
//  ====================================================================================================                     
// My modifications - Thanks to HDM's great work
// tornado@odessaua.com
//
// New for v2.0 -Added 13 new Flash Chip Parts to the list:
//                     - Macronix MX25L160A         (2MB)
//                     - W19B(L)320ST   2Mx16 TopB  (4MB)
//                     - W19B(L)320SB   2Mx16 BotB  (4MB)
//                     - M29DW324DT 2Mx16 TopB      (4MB)
//                     - M29DW324DB 2Mx16 BotB      (4MB)
//                     - TC58FVM6T2A  4Mx16 TopB    (8MB)
//                     - TC58FVM6B2A  4Mx16 BopB    (8MB)
//                     - K8D6316UTM  4Mx16 TopB     (8MB)
//                     - K8D6316UBM  4Mx16 BotB     (8MB)
//                     - Spansion S29GL032M BotB    (4MB)
//                     - Spansion S29GL032M TopB    (4MB)
//                     - Spansion S29GL064M BotB    (8MB)
//                     - Spansion S29GL064M TopB    (8MB)
//                 Added 2 New Processors
//                     - Broadcom BCM4785 Rev 1 - WRT350N
//                     - TI TNETV1060GDW
//                 Added Spansion 3 Byte Device ID Workaround
//                       Spansion Flash & TI TNETV1060GDW CPU
//                       Contributed by kb1klk@aetheriatech.org
//		   Added Wiggler.png - tjtag works with this clone
//			 tested Tornado
//		   Added Changelog.txt, and removed from source.		
//
// New for v4.8.1 -Added 2 new Flash Chip Parts to the list:
//                     - EN29LV160A  2Mx16 TopB     (2MB)
//                     - EN29LV160A  2Mx16 BotB     (2MB)
//                 Added 4 New Processors
//                     - BCM5453 Rev 1
//                     - BCM5453 Rev 2
//                     - BCM5365 Rev 1
//                     - BCM4704 Rev 9
//                 Added the following New Switch Options
//                     - /bypass .............Unlock Bypass
//                       for AMD/Spansion type Flash Chips
//                 Added BSP option for use with flash, erase & backup
//
//  New for v4.8 - Added 2 new Flash Chip Parts to the list:
//                     - SST39VF6401B 4Mx16 BotB     (8MB)
//                     - SST39VF6402B 4Mx16 TopB     (8MB)
//               - Added the following New Switch Options
//                     - /wiggler ........... use wiggler cable
//
//  New for v4.7 - Added 2 new Flash Chip Parts to the list:
//                     - K8D3216UTC  2Mx16 TopB     (4MB)
//                     - K8D3216UBC  2Mx16 BotB     (4MB)
//
//  New for v4.6 - Added Common Flash Chip Polling routine
//               - Added "-probeonly" parameter (good idea jmranger)
//               - Added Chip ID for Broadcom BCM4704 Rev 8 CPU
//               - Added TRST Signal Support for Wiggler Cables
//               - Added Chip ID for BRECIS MSP2007-CA-A1 CPU
//               - Added Experimental 1MB Flash Chip Offsets
//               - Added 2 new Flash Chip Parts to the list:
//                     - MX29LV800BTC 512kx16 TopB  (1MB)
//                     - MX29LV800BTC 512kx16 BotB  (1MB)
//
//  New for v4.5 - Added 2 new Flash Chip Parts to the list:
//                     - K8D1716UTC 1Mx16 TopB      (2MB)
//                     - K8D1716UBC 1Mx16 BotB      (2MB)
//
//  New for v4.4 - Added PrAcc routines to support additional MIPS chips
//                 without the ability to use EJTAG DMA Access
//               - Added Chip ID for Broadcom BCM5365 Rev 1 CPU
//               - Added Chip ID for Broadcom BCM6348 Rev 1 CPU (Big Endian)
//               - Added Chip ID for Broadcom BCM6345 Rev 1 CPU
//               - Added 6 new Flash Chip Parts to the list:
//                     - SST39VF1601 1Mx16 BotB     (2MB)
//                     - SST39VF1602 1Mx16 TopB     (2MB)
//                     - SST39VF3201 2Mx16 BotB     (4MB)
//                     - SST39VF3202 2Mx16 TopB     (4MB)
//                     - SST39VF6401 4Mx16 BotB     (8MB)
//                     - SST39VF6402 4Mx16 TopB     (8MB)
//               - Added the following New Switch Options
//                     - /noemw ............. prevent Enabling Memory Writes
//                     - /nocwd ............. prevent Clearing CPU Watchdog Timer
//                     - /dma ............... force use of DMA routines
//                     - /nodma ............. force use of PRACC routines (No DMA)
//                     - /window:XXXXXXXX ... custom flash window base (in HEX)
//                     - /start:XXXXXXXX .... custom start location (in HEX)
//                     - /length:XXXXXXXX ... custom length (in HEX)
//                     - /silent ............ prevent scrolling display of data
//                     - /skipdetect ........ skip auto detection of CPU Chip ID
//                     - /instrlen:XX ....... set instruction length manually
//               - Added elapsed time to Backup, Erase, and Flash routines
//               - Other minor miscellaneous changes/additions.
//
//  New for v4.3 - Corrected Macronix Flash Chip Block Definitions.
//               - Add 8 new Flash Chip Parts to the list:
//                     - AT49BV/LV16X 2Mx16 BotB    (4MB)
//                     - AT49BV/LV16XT 2Mx16 TopB   (4MB)
//                     - MBM29LV160B 1Mx16 BotB     (2MB)
//                     - MBM29LV160T 1Mx16 TopB     (2MB)
//                     - MX29LV161B 1Mx16 BotB      (2MB)
//                     - MX29LV161T 1Mx16 TopB      (2MB)
//                     - ST M29W160EB 1Mx16 BotB    (2MB)
//                     - ST M29W160ET 1Mx16 TopB    (2MB)
//
//  New for v4.2 - Changed the chip_detect routine to allow for easier
//                 additions of new chip id's.
//               - Added detection support for the Broadcom BCM5350 chip.
//               - Fixed DMA routines to check status bit that was
//                 removed in prior version.
//               - Removed clockout routine in an effort to speed up access.
//               - Changed clockin routine in an effort to speed up access.
//               - Changed ReadData and WriteData routines to merely call
//                 ReadWriteData routine.
//               - Removed Defines from .h file and placed flash areas in a
//                 structure list for easier maintenance should they change.
//               - Miscellaneous other minor changes.
//
// **************************************************************************
//  Written by HairyDairyMaid (a.k.a. - lightbulb)
//  hairydairymaid@yahoo.com
// **************************************************************************
//
//  This program is copyright (C) 2004 HairyDairyMaid (a.k.a. Lightbulb)
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of version 2 the GNU General Public License as published
//  by the Free Software Foundation.
//  This program is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
//  more details.
//  To view a copy of the license go to:
//  http://www.fsf.org/copyleft/gpl.html
//  To receive a copy of the GNU General Public License write the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// **************************************************************************

#define WINDOWS_VERSION   // uncomment this for Windows Compile 
                          // MS Visual C Compiler: cl brjtag.c

#ifdef WINDOWS_VERSION

#include <windows.h>      // Only for Windows Compile
#define strcasecmp  stricmp
#define strncasecmp strnicmp
#define _OUTP(d) _outp(lpt_port, d)
#define _INP(d)   d = (unsigned char)_inp(lpt_port+1)
#define tnano(s)     Sleep((s) / 1000000)
#define mysleep(s)   Sleep((s) * 1000)
/* Windows sleep is milliseconds, time/1000000 gives us nanoseconds */

#else   //For linux

#define _OUTP(d)  ioctl(pfd, PPWDATA, &d)
#define _INP(d)		ioctl(pfd, PPRSTATUS, &d)
#define tnano(s)     sleep((s) / 1000000000)
#define mysleep(s)   sleep(s)

#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "brjtag.h"
#define BigEndian		endian
#define DEBUGMSG1 0
#define DEBUGMSG  0

static unsigned int ctrl_reg;

int pfd;
int lpt_port	       = PORT378;
int instruction_length;
int issue_reset      = 1;
int issue_enable_mw  = 1;
int issue_watchdog   = 1;
int issue_break      = 1;
int issue_erase      = 1;
int issue_timestamp  = 1;
int force_dma        = 0;
int force_nodma      = 0;
int selected_fc      = 0;
unsigned int selected_window  = 0;
unsigned int selected_start   = 0;
unsigned int selected_length  = 0;
unsigned int * init_code = NULL;
int custom_options   = 0;
int silent_mode      = 1;
int skipdetect       = 0;
int instrlen         = 0;
int cable_type       = XILINX;
int endian     	     = END_MIPSLE;
int force_rev_endian = 0;
int force_tinycfe    = 0;
int force_align      = 0;
int issue_initcpu    = 0;
int issue_chip_erase = 0;
int issue_clear_ppb  = 0;
int issue_show_ppb   = 0;
int PPB_type         = 0;
int probeme          = 0;   
int USE_DMA          = 0;
int ejtag_version    = 0;
int bypass           = 0;

char            flash_part[128];
unsigned int    flash_size = 0;

int             block_total = 0;
unsigned int    block_addr = 0;
unsigned int    blocks[1025];
unsigned int    cmd_type = 0;

char            AREA_NAME[128];
unsigned int    AREA_START;
unsigned int    AREA_LENGTH;
unsigned int    FLASH_MEMORY_START;
unsigned int    FLASH_PROBE_BASE;
unsigned int    vendid;
unsigned int    devid;

unsigned int    data_register;
unsigned int    address_register;
int justwrite = true;



typedef struct _processor_chip_type
  {
    unsigned int        chip_id;        // Processor Chip ID
    int                 instr_length;   // EJTAG Instruction Length
    int									chip_endn;  		// Chip endian
    unsigned int *      init_code;      // Initialize code start 
    char*               chip_descr;     // Processor Chip Description
    
  } processor_chip_type;

processor_chip_type  processor_chip_list[] =
{
/* ********* Broadcom Wifi Chips *************/    
  { 0x0471017F, 5, END_MIPSLE, NULL, "Broadcom BCM4702 Rev 1 CPU" },
  { 0x9470417F, 8, END_MIPSLE, NULL, "Broadcom BCM4704 KPBG Rev 9 CPU"},     // Tornado  WRT150N
  { 0x0470417F, 8, END_MIPSLE, NULL, "Broadcom BCM4704 Rev 8 CPU" },         // BCM4704 chip (used in the WRTSL54GS units)
  { 0x1471217F, 8, END_MIPSLE, NULL, "Broadcom BCM4712 Rev 1 CPU" },
  { 0x2471217F, 8, END_MIPSLE, NULL, "Broadcom BCM4712 Rev 2 CPU" },
  { 0x0478517F, 8, END_MIPSLE, NULL, "Broadcom BCM4785 Rev 1 CPU" },         // Tornado WRT350N
  { 0x0535017F, 8, END_MIPSLE, NULL, "Broadcom BCM5350 Rev 1 CPU" },
  { 0x0535217F, 8, END_MIPSLE, NULL, "Broadcom BCM5352 Rev 1 CPU" },
  { 0x1535417F, 8, END_MIPSLE, NULL, "Broadcom BCM5354 KFBG Rev 1 CPU" }, 	 // Tornado - WRT54G GV8/GSV7
  { 0x2535417F, 8, END_MIPSLE, NULL, "Broadcom BCM5354 KFBG Rev 2 CPU" },    // Tornado - Gv8/GSv7
  { 0x0536517F, 8, END_MIPSLE, NULL, "Broadcom BCM5365 Rev 1 CPU" },         // BCM5365 Not Completely Verified Yet
  { 0x1536517F, 8, END_MIPSLE, NULL, "Broadcom BCM5365 Rev 1 CPU" },         // Eko....ASUS WL 500 G Deluxe
/* ********* Broadcom ADSL Chips *************/  
  { 0x0634517F, 5, END_MIPS,   NULL, "Broadcom BCM6345 Rev 1 CPU" },         // BCM6345 Not Completely Verified Yet
  { 0x0633817F, 5, END_MIPS,   NULL, "Broadcom BCM6338 Rev 1 CPU" },         //Fully Tested
  { 0x0634817F, 5, END_MIPS,   NULL, "Broadcom BCM6348 Rev 1 CPU" },         //Fully Tested
  { 0x0635817F, 5, END_MIPS,   init_6358, "Broadcom BCM6358 Rev 1 CPU" },    //Fully Tested
  { 0x0636817F, 5, END_MIPS,   NULL, "Broadcom BCM6368 Rev 1 CPU" },
/* ********* Non Broadcom Chips *************/
  { 0x0000100F, 5, END_MIPS,   NULL, "TI AR7WRD TNETD7300GDU Rev 1 CPU" },   // TI AR7WRD Only Partially Verified
  { 0x102002E1, 5, END_MIPSLE, NULL, "BRECIS MSP2007-CA-A1 CPU" },           // BRECIS chip - Not Completely Verified Yet
  { 0x0B52D02F, 5, END_MIPSLE, NULL, "TI TNETV1060GDW CPU" },                // Fox WRTP54G
/* ****** Add your self CPU before this line ****/
  { 0, 0, 0, 0 }
};


typedef struct _flash_area_type
  {
    unsigned int        chip_size;
    char*               area_name;
    unsigned int        area_start;
    unsigned int        area_length;
  } flash_area_type;


flash_area_type  flash_area_list[] =
{
  //---------   ----------     -----------  ------------
  //chip_size   area_name      area_start   area_length
  //---------   ----------     -----------  ------------
  { size1MB,    "CFE",         0x1FC00000,  0x40000 },
  { size2MB,    "CFE",         0x1FC00000,  0x40000 },
  { size4MB,    "CFE",         0x1FC00000,  0x40000 },
  { size8MB,    "CFE",         0x1C000000,  0x40000 },
  { size16MB,   "CFE",         0x1C000000,  0x40000 },
  // for tiny CFE
  { size1MB,    "TFE",         0x1FC00000,  0x10000 },
  { size2MB,    "TFE",         0x1FC00000,  0x10000 },
  { size4MB,    "TFE",         0x1FC00000,  0x10000 },    // Tiny CFE has one bottom sector size,
  { size8MB,    "TFE",         0x1F800000,  0x20000 },    // Program will re-calculate the address for ADSL chip.
  { size16MB,   "TFE",         0x1F000000,  0x20000 },

  { size1MB,    "KERNEL",      0x1FC40000,  0x0B0000 },
  { size2MB,    "KERNEL",      0x1FC40000,  0x1B0000 },
  { size4MB,    "KERNEL",      0x1FC40000,  0x3B0000 },
  { size8MB,    "KERNEL",      0x1C040000,  0x7A0000 },
  { size16MB,   "KERNEL",      0x1C040000,  0x7A0000 },
// for tiny CFE
  { size1MB,    "TERNEL",      0x1FC10000,  0x0E0000 },
  { size2MB,    "TERNEL",      0x1FC10000,  0x1E0000 },
  { size4MB,    "TERNEL",      0x1FC10000,  0x3E0000 },    //vs. Tiny CFE
  { size8MB,    "TERNEL",      0x1F810000,  0x7D0000 },
  { size16MB,   "TERNEL",      0x1F010000,  0x7D0000 },

  { size1MB,    "NVRAM",       0x1FCF0000,  0x10000 },
  { size2MB,    "NVRAM",       0x1FDF0000,  0x10000 },
  { size4MB,    "NVRAM",       0x1FFF0000,  0x10000 },
  { size8MB,    "NVRAM",       0x1C7E0000,  0x20000 },
  { size16MB,   "NVRAM",       0x1C7E0000,  0x20000 },

  { size1MB,    "WHOLEFLASH",  0x1FC00000,  0x100000 },
  { size2MB,    "WHOLEFLASH",  0x1FC00000,  0x200000 },
  { size4MB,    "WHOLEFLASH",  0x1FC00000,  0x400000 },
  { size8MB,    "WHOLEFLASH",  0x1C000000,  0x800000 },
  { size16MB,   "WHOLEFLASH",  0x1C000000, 0x1000000 },

  { size1MB,    "BSP",         0x1FC00000,  0x50000 },
  { size2MB,    "BSP",         0x1FC00000,  0x50000 },
  { size4MB,    "BSP",         0x1FC00000,  0x50000 },
  { size8MB,    "BSP",         0x1C000000,  0x50000 },
  { size16MB,   "BSP",         0x1C000000,  0x50000 },


  { 0, 0, 0, 0 }
};


typedef struct _flash_chip_type
  {
    unsigned int        vendid;         // Manufacturer Id
    unsigned int        devid;          // Device Id
    unsigned int        flash_size;     // Total size in MBytes
    unsigned int        cmd_type;       // Device CMD TYPE
    unsigned int        prot_type;      // Block Protection Type, 
    char*               flash_part;     // Flash Chip Description
    unsigned int        region1_num;    // Region 1 block count
    unsigned int        region1_size;   // Region 1 block size
    unsigned int        region2_num;    // Region 2 block count
    unsigned int        region2_size;   // Region 2 block size
    unsigned int        region3_num;    // Region 3 block count
    unsigned int        region3_size;   // Region 3 block size
    unsigned int        region4_num;    // Region 4 block count
    unsigned int        region4_size;   // Region 4 block size
    unsigned int        region5_num;    // Region 5 block count
    unsigned int        region5_size;   // Region 5 block size
  } flash_chip_type;


flash_chip_type  flash_chip_list[] =
{
  /* AMD, Spansion */
  { 0x00C2, 0x22DA, size1MB, CMD_TYPE_AMD, 0, "MX29LV800BTC 512kx16 TopB  (1MB)"   ,15,size32K,    1,size16K,    2,size4K,   1,size8K  },
  { 0x00C2, 0x225B, size1MB, CMD_TYPE_AMD, 0, "MX29LV800BTC 512kx16 BotB  (1MB)"   ,1,size8K,      2,size4K,     1,size16K,  15,size32K},

  { 0x0001, 0x2249, size2MB, CMD_TYPE_AMD, 0, "AMD 29lv160DB 1Mx16 BotB   (2MB)"   ,1,size16K,    2,size8K,     1,size32K,  31,size64K },
  { 0x0001, 0x22c4, size2MB, CMD_TYPE_AMD, 0, "AMD 29lv160DT 1Mx16 TopB   (2MB)"   ,31,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x007f, 0x2249, size2MB, CMD_TYPE_AMD, 0, "EON EN29LV160A 1Mx16 BotB  (2MB)"   ,1,size16K,    2,size8K,     1,size32K,  31,size64K },
  { 0x007f, 0x22C4, size2MB, CMD_TYPE_AMD, 0, "EON EN29LV160A 1Mx16 TopB  (2MB)"   ,31,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x0004, 0x2249, size2MB, CMD_TYPE_AMD, 0, "MBM29LV160B 1Mx16 BotB     (2MB)"   ,1,size16K,    2,size8K,     1,size32K,  31,size64K },
  { 0x0004, 0x22c4, size2MB, CMD_TYPE_AMD, 0, "MBM29LV160T 1Mx16 TopB     (2MB)"   ,31,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x00C2, 0x2249, size2MB, CMD_TYPE_AMD, 0, "MX29LV161B 1Mx16 BotB      (2MB)"   ,1,size16K,    2,size8K,     1,size32K,  31,size64K },
  { 0x00C2, 0x22c4, size2MB, CMD_TYPE_AMD, 0, "MX29LV161T 1Mx16 TopB      (2MB)"   ,31,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x00EC, 0x2275, size2MB, CMD_TYPE_AMD, 0, "K8D1716UTC  1Mx16 TopB     (2MB)"   ,31,size64K,    8,size8K,     0,0,        0,0       },
  { 0x00EC, 0x2277, size2MB, CMD_TYPE_AMD, 0, "K8D1716UBC  1Mx16 BotB     (2MB)"   ,8,size8K,      31,size64K,   0,0,        0,0       },
  { 0x0020, 0x2249, size2MB, CMD_TYPE_AMD, 0, "ST M29W160EB 1Mx16 BotB    (2MB)"   ,1,size16K,    2,size8K,     1,size32K,  31,size64K },
  { 0x0020, 0x22c4, size2MB, CMD_TYPE_AMD, 0, "ST M29W160ET 1Mx16 TopB    (2MB)"   ,31,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x00C2, 0x0014, size2MB, CMD_TYPE_AMD, 0, "Macronix MX25L160A         (2MB)"   ,32,size64K,   0,0,          0,0,        0,0        }, /* new */

  { 0x00EC, 0x22A0, size2MB, CMD_TYPE_AMD, 0, "K8D3216UTC  2Mx16 TopB     (4MB)"   ,63,size64K,    8,size8K,     0,0,        0,0       },
  { 0x00EC, 0x22A2, size2MB, CMD_TYPE_AMD, 0, "K8D3216UBC  2Mx16 BotB     (4MB)"   ,8,size8K,      63,size64K,   0,0,        0,0       },

  { 0x0001, 0x2200, size4MB, CMD_TYPE_AMD, 0, "AMD 29lv320MB 2Mx16 BotB   (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x0001, 0x227E, size4MB, CMD_TYPE_AMD, 0, "AMD 29lv320MT 2Mx16 TopB   (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0001, 0x2201, size4MB, CMD_TYPE_AMD, 0, "AMD 29lv320MT 2Mx16 TopB   (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0098, 0x009C, size4MB, CMD_TYPE_AMD, 0, "TC58FVB321 2Mx16 BotB      (4MB)"   ,1,size16K,    2,size8K,     1,size32K,  63,size64K },
  { 0x0098, 0x009A, size4MB, CMD_TYPE_AMD, 0, "TC58FVT321 2Mx16 TopB      (4MB)"   ,63,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x001F, 0x00C0, size4MB, CMD_TYPE_AMD, 0, "AT49BV/LV16X 2Mx16 BotB    (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x001F, 0x00C2, size4MB, CMD_TYPE_AMD, 0, "AT49BV/LV16XT 2Mx16 TopB   (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0004, 0x2253, size4MB, CMD_TYPE_AMD, 0, "MBM29DL323BE 2Mx16 BotB    (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x0004, 0x2250, size4MB, CMD_TYPE_AMD, 0, "MBM29DL323TE 2Mx16 TopB    (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0001, 0x22f9, size4MB, CMD_TYPE_AMD, 0, "AMD 29lv320DB 2Mx16 BotB   (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x0001, 0x22f6, size4MB, CMD_TYPE_AMD, 0, "AMD 29lv320DT 2Mx16 TopB   (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0004, 0x22F9, size4MB, CMD_TYPE_AMD, 0, "MBM29LV320BE 2Mx16 BotB    (4MB)"   ,1,size16K,    2,size8K,     1,size32K,  63,size64K },
  { 0x0004, 0x22F6, size4MB, CMD_TYPE_AMD, 0, "MBM29LV320TE 2Mx16 TopB    (4MB)"   ,63,size64K,   1,size32K,    2,size8K,   1,size16K  },
  { 0x00C2, 0x22A8, size4MB, CMD_TYPE_AMD, 0, "MX29LV320AB 2Mx16 BotB     (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x00C2, 0x00A8, size4MB, CMD_TYPE_AMD, 0, "MX29LV320AB 2Mx16 BotB     (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x00C2, 0x00A7, size4MB, CMD_TYPE_AMD, 0, "MX29LV320AT 2Mx16 TopB     (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x00C2, 0x22A7, size4MB, CMD_TYPE_AMD, 0, "MX29LV320AT 2Mx16 TopB     (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0020, 0x22CB, size4MB, CMD_TYPE_AMD, 0, "ST 29w320DB 2Mx16 BotB     (4MB)"   ,1,size16K,    2,size8K,     1,size32K,  63,size64K },
  { 0x0020, 0x22CA, size4MB, CMD_TYPE_AMD, 0, "ST 29w320DT 2Mx16 TopB     (4MB)"   ,63,size64K,   1,size32K,    2,size8K,   1,size16K  },

  { 0x00DA, 0x22BA, size4MB, CMD_TYPE_AMD, 0, "W19B(L)320ST   2Mx16 TopB  (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        }, /* new */
  { 0x00DA, 0x222A, size4MB, CMD_TYPE_AMD, 0, "W19B(L)320SB   2Mx16 BotB  (4MB)"   ,8,size8K,   63,size64K,     0,0,        0,0        }, /* new */

  { 0x0020, 0x225C, size4MB, CMD_TYPE_AMD, 0, "M29DW324DT 2Mx16 TopB      (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        }, /* new */
  { 0x0020, 0x225D, size4MB, CMD_TYPE_AMD, 0, "M29DW324DB 2Mx16 BotB      (4MB)"   ,8,size8K,   63,size64K,     0,0,        0,0        }, /* new */

  { 0x0098, 0x0057, size8MB, CMD_TYPE_AMD, 0, "TC58FVM6T2A  4Mx16 TopB    (8MB)"   ,127,size64K,   8,size8K,     0,0,        0,0       }, /* new */
  { 0x0098, 0x0058, size8MB, CMD_TYPE_AMD, 0, "TC58FVM6B2A  4Mx16 BopB    (8MB)"   ,8,size8K,   127,size64K,     0,0,        0,0       }, /* new */

  { 0x00EC, 0x22E0, size8MB, CMD_TYPE_AMD, 0, "K8D6316UTM  4Mx16 TopB     (8MB)"   ,127,size64K,    8,size8K,     0,0,        0,0      }, /* new */
  { 0x00EC, 0x22E2, size8MB, CMD_TYPE_AMD, 0, "K8D6316UBM  4Mx16 BotB     (8MB)"   ,8,size8K,      127,size64K,   0,0,        0,0      }, /* new */

  /* BSC */
  { 0x0089, 0x8891, size2MB, CMD_TYPE_BSC, 0, "Intel 28F160B3 1Mx16 BotB  (2MB)"   ,8,size8K,     31,size64K,   0,0,        0,0        },
  { 0x0089, 0x8890, size2MB, CMD_TYPE_BSC, 0, "Intel 28F160B3 1Mx16 TopB  (2MB)"   ,31,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0089, 0x88C3, size2MB, CMD_TYPE_BSC, 0, "Intel 28F160C3 1Mx16 BotB  (2MB)"   ,8,size8K,     31,size64K,   0,0,        0,0        },
  { 0x0089, 0x88C2, size2MB, CMD_TYPE_BSC, 0, "Intel 28F160C3 1Mx16 TopB  (2MB)"   ,31,size64K,   8,size8K,     0,0,        0,0        },

  { 0x0089, 0x8897, size4MB, CMD_TYPE_BSC, 0, "Intel 28F320B3 2Mx16 BotB  (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x0089, 0x8896, size4MB, CMD_TYPE_BSC, 0, "Intel 28F320B3 2Mx16 TopB  (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x0089, 0x88C5, size4MB, CMD_TYPE_BSC, 0, "Intel 28F320C3 2Mx16 BotB  (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x0089, 0x88C4, size4MB, CMD_TYPE_BSC, 0, "Intel 28F320C3 2Mx16 TopB  (4MB)"   ,63,size64K,   8,size8K,     0,0,        0,0        },
  { 0x00b0, 0x00e3, size4MB, CMD_TYPE_BSC, 0, "Sharp 28F320BJE 2Mx16 BotB (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },

  { 0x0089, 0x8899, size8MB, CMD_TYPE_BSC, 0, "Intel 28F640B3 4Mx16 BotB  (8MB)"   ,8,size8K,     127,size64K,  0,0,        0,0        },
  { 0x0089, 0x8898, size8MB, CMD_TYPE_BSC, 0, "Intel 28F640B3 4Mx16 TopB  (8MB)"   ,127,size64K,  8,size8K,     0,0,        0,0        },
  { 0x0089, 0x88CD, size8MB, CMD_TYPE_BSC, 0, "Intel 28F640C3 4Mx16 BotB  (8MB)"   ,8,size8K,     127,size64K,  0,0,        0,0        },
  { 0x0089, 0x88CC, size8MB, CMD_TYPE_BSC, 0, "Intel 28F640C3 4Mx16 TopB  (8MB)"   ,127,size64K,  8,size8K,     0,0,        0,0        },

  /* SCS */
  { 0x00b0, 0x00d0, size2MB, CMD_TYPE_SCS, 0, "Intel 28F160S3/5 1Mx16     (2MB)"   ,32,size64K,   0,0,          0,0,        0,0        },

  { 0x0089, 0x0016, size4MB, CMD_TYPE_SCS, 0, "Intel 28F320J3 2Mx16       (4MB)"   ,32,size128K,  0,0,          0,0,        0,0        },
  { 0x0089, 0x0014, size4MB, CMD_TYPE_SCS, 0, "Intel 28F320J5 2Mx16       (4MB)"   ,32,size128K,  0,0,          0,0,        0,0        },
  { 0x00b0, 0x00d4, size4MB, CMD_TYPE_SCS, 0, "Intel 28F320S3/5 2Mx16     (4MB)"   ,64,size64K,   0,0,          0,0,        0,0        },

  { 0x0089, 0x0017, size8MB, CMD_TYPE_SCS, 0, "Intel 28F640J3 4Mx16       (8MB)"   ,64,size128K,  0,0,          0,0,        0,0        },
  { 0x0089, 0x0015, size8MB, CMD_TYPE_SCS, 0, "Intel 28F640J5 4Mx16       (8MB)"   ,64,size128K,  0,0,          0,0,        0,0        },

  { 0x0089, 0x0018, size16MB, CMD_TYPE_SCS, 0, "Intel 28F128J3 8Mx16      (16MB)"  ,128,size128K, 0,0,          0,0,        0,0        },

  /* SST */
  { 0x00BF, 0x234B, size4MB, CMD_TYPE_SST, 0, "SST39VF1601 1Mx16 BotB     (2MB)"   ,64,size32K,    0,0,          0,0,        0,0       },
  { 0x00BF, 0x234A, size4MB, CMD_TYPE_SST, 0, "SST39VF1602 1Mx16 TopB     (2MB)"   ,64,size32K,    0,0,          0,0,        0,0       },

  { 0x00BF, 0x235B, size4MB, CMD_TYPE_SST, 0, "SST39VF3201 2Mx16 BotB     (4MB)"   ,128,size32K,   0,0,          0,0,        0,0       },
  { 0x00BF, 0x235A, size4MB, CMD_TYPE_SST, 0, "SST39VF3202 2Mx16 TopB     (4MB)"   ,128,size32K,   0,0,          0,0,        0,0       },
  { 0x00BF, 0x2783, size4MB, CMD_TYPE_SST, 0, "SST39VF320 2Mx16           (4MB)"   ,64,size64K,   0,0,          0,0,        0,0        },

  { 0x00BF, 0x236B, size4MB, CMD_TYPE_SST, 0, "SST39VF6401 4Mx16 BotB     (8MB)"   ,256,size32K,   0,0,          0,0,        0,0       },
  { 0x00BF, 0x236A, size4MB, CMD_TYPE_SST, 0, "SST39VF6402 4Mx16 TopB     (8MB)"   ,256,size32K,   0,0,          0,0,        0,0       },
  { 0x00BF, 0x236D, size4MB, CMD_TYPE_SST, 0, "SST39VF6401B 4Mx16 BotB    (8MB)"   ,256,size32K,   0,0,          0,0,        0,0       },
  { 0x00BF, 0x236C, size4MB, CMD_TYPE_SST, 0, "SST39VF6402B 4Mx16 TopB    (8MB)"   ,256,size32K,   0,0,          0,0,        0,0       },

//  See Spansion hack details for reasoning for the unusual vendid
  { 0x017E, 0x1A00, size4MB, CMD_TYPE_AMD, 0, "Spansion S29GL032MR4 BotB  (4MB)"   ,8,size8K,     63,size64K,   0,0,        0,0        },
  { 0x017E, 0x1A01, size4MB, CMD_TYPE_AMD, 0, "Spansion S29GL032MR3 TopB  (4MB)"   ,63,size64K,     8,size8K,   0,0,        0,0        },
  { 0x017E, 0x1000, size8MB, CMD_TYPE_AMD, 0, "Spansion S29GL064MR4 BotB  (8MB)"   ,8,size8K,     127,size64K,   0,0,        0,0       },
  { 0x017E, 0x1001, size8MB, CMD_TYPE_AMD, 0, "Spansion S29GL064MR3 TopB  (8MB)"   ,127,size64K,     8,size8K,   0,0,        0,0       },

//  Spansion uniform sector flash
  { 0x017E, 0x1200, size16MB, CMD_TYPE_AMD, 0, "Spansion S29GL128M Uni    (16MB)"   ,128,size128K,   0,0,      0,0,        0,0         },
  { 0x017E, 0x2101, size16MB, CMD_TYPE_AMD, SP_PPB, "Spansion S29GL128N/P Uni  (16MB)",128,size128K,   0,0,      0,0,        0,0       },

  { 0x017E, 0x1300, size8MB, CMD_TYPE_AMD, 0, "Spansion S29GL064MR0 Uni      (8MB)" ,128,size64K,    0,0,      0,0,        0,0         },
  { 0x017E, 0x0C01, size8MB, CMD_TYPE_AMD, 0, "Spansion S29GL064MR1,2,8,9 Uni(8MB)" ,128,size64K,    0,0,      0,0,        0,0         },
  { 0x017E, 0x1301, size8MB, CMD_TYPE_AMD, 0, "Spansion S29GL064MR5,6,7 Uni  (8MB)" ,128,size64K,    0,0,      0,0,        0,0         },
  
  { 0x017E, 0x1C00, size4MB, CMD_TYPE_AMD, 0, "Spansion S29GL032MR0 Uni   (4MB)"    ,64,size64K,    0,0,      0,0,        0,0          },
  { 0x017E, 0x1D00, size4MB, CMD_TYPE_AMD, 0, "Spansion S29GL032MR1,2 Uni (4MB)"    ,64,size64K,    0,0,      0,0,        0,0          },
  
// Add new flash model before this line  
  { 0, 0, 0, 0, 0,   0,0,   0,0,   0,0,   0,0 }
};


// -----------------------------------------
// ---- Start of Compiler Specific Code ----
// -----------------------------------------

void lpt_openport(void)
{
#ifdef WINDOWS_VERSION    // ---- Compiler Specific Code ----

  HANDLE h;

  h = CreateFile("\\\\.\\giveio", GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (h == INVALID_HANDLE_VALUE)
    {
      printf("Couldn't access giveio device\n");
      CloseHandle(h);
      exit(0);
    }
  CloseHandle(h);

#else                     // ---- Compiler Specific Code ----

#ifdef __FreeBSD__     // ---- Compiler Specific Code ----

  pfd = open("/dev/ppi0", O_RDWR);
  if (pfd < 0)
    {
      perror("Failed to open /dev/ppi0");
      exit(0);
    }
  if ((ioctl(pfd, PPEXCL) < 0) || (ioctl(pfd, PPCLAIM) < 0))
    {
      perror("Failed to lock /dev/ppi0");
      close(pfd);
      exit(0);
    }

#else                  // ---- Compiler Specific Code ----

  pfd = open("/dev/parport0", O_RDWR);
  if (pfd < 0)
    {
      perror("Failed to open /dev/parport0");
      exit(0);
    }
  if ((ioctl(pfd, PPEXCL) < 0) || (ioctl(pfd, PPCLAIM) < 0))
    {
      perror("Failed to lock /dev/parport0");
      close(pfd);
      exit(0);
    }

#endif

#endif
}


void lpt_closeport(void)
{
#ifndef WINDOWS_VERSION   // ---- Compiler Specific Code ----

#ifndef __FreeBSD__    // ---- Compiler Specific Code ----

  if (ioctl(pfd, PPRELEASE) < 0)
    {
      perror("Failed to release /dev/parport0");
      close(pfd);
      exit(0);
    }

#endif

  close(pfd);

#endif
}


static unsigned char clockin(int tms, int tdi)
{
  unsigned char data;
 
  tms = tms ? 1 : 0;
  tdi = tdi ? 1 : 0;

 switch( cable_type )
    {
        case WIGGLER:
           data = (1 << WTDO) | (0 << WTCK) | (tms << WTMS) | (tdi << WTDI) | (1 << WTRST_N);
           _OUTP(data);
           data = (1 << WTDO) | (1 << WTCK) | (tms << WTMS) | (tdi << WTDI) | (1 << WTRST_N);
           _OUTP(data);
           _INP(data);          // output inversed bit 7
           data ^= 0x80;        // invert bit 7
           data >>= WTDO;       // shift bit 7 to bit 0
           data &= 1;           // get bit 0 output
           break;      
           
 				default:    //XILINX
           data = (1 << TDO) | (0 << TCK) | (tms << TMS) | (tdi << TDI);
           _OUTP(data);
           data = (1 << TDO) | (1 << TCK) | (tms << TMS) | (tdi << TDI);
           _OUTP(data);
           _INP(data);          // output bit 4
//           data ^= 0x80;        // invert bit 7
           data >>= TDO;        // shift bit 4 to bit 0 
           data &= 1;           // get bit 0 output
     }
     
  return data;
}


// ---------------------------------------
// ---- End of Compiler Specific Code ----
// ---------------------------------------

void test_reset(void)
{
  clockin(1, 0);  // Run through a handful of clock cycles with TMS high to make sure
  clockin(1, 0);  // we are in the TEST-LOGIC-RESET state.
  clockin(1, 0);
  clockin(1, 0);
  clockin(1, 0);
  clockin(0, 0);  // enter runtest-idle
}


void set_instr(int instr)
{
  int i;
  static int curinstr = 0xFFFFFFFF;

  if (instr == curinstr)
    return;

  clockin(1, 0);  // enter select-dr-scan
  clockin(1, 0);  // enter select-ir-scan
  clockin(0, 0);  // enter capture-ir
  clockin(0, 0);  // enter shift-ir (dummy)
  for (i=0; i < instruction_length; i++)
    {
      clockin(i==(instruction_length - 1), (instr>>i)&1);
    }
  clockin(1, 0);  // enter update-ir
  clockin(0, 0);  // enter runtest-idle

  curinstr = instr;
}


static unsigned int ReadWriteData(unsigned int in_data)
{
  int i;
  unsigned int out_data = 0;
  unsigned char out_bit;

  clockin(1, 0);  // enter select-dr-scan
  clockin(0, 0);  // enter capture-dr
  clockin(0, 0);  // enter shift-dr
  for (i = 0 ; i < 32 ; i++)
    {
      out_bit  = clockin((i == 31), ((in_data >> i) & 1)); //final bit 1 exit-dr
      out_data = out_data | (out_bit << i);
    }
  clockin(1,0);   // enter update-dr
  clockin(0,0);   // enter runtest-idle
  return out_data;
}


static unsigned int ReadData(void)
{
  return ReadWriteData(0x00);
}


void WriteData(unsigned int in_data)
{
  ReadWriteData(in_data);
}


void ShowData(unsigned int value)
{
  int i;
  for (i=0; i<32; i++)
    printf("%d", (value >> (31-i)) & 1);
  printf(" (%08X)\n", value);
}

void ShowData_h(unsigned int value)
{
  int i;
  value = value & 0xFFFF;
  for (i=0; i<16; i++)
  printf("%d", (value >> (16-i)) & 1);
  printf(" (%04X)\n", value);
}


//memory(flash) read always with word format
static unsigned int ejtag_read(unsigned int addr)
{
  if (USE_DMA) return(ejtag_dma_read(addr));
  else return(ejtag_pracc_read2(addr));
}

//Use for setting flash command
static unsigned int ejtag_read_h(unsigned int addr)
{
  if (USE_DMA) return(ejtag_dma_read_h(addr));
  else return(ejtag_pracc_read_h2(addr));
}

//word write, only used for DCR setting
void ejtag_write(unsigned int addr, unsigned int data)
{
  if (USE_DMA) ejtag_dma_write(addr, data);
  else ejtag_pracc_write2(addr, data);
}

//all flash operations are in 16bits(half-word)
void ejtag_write_h(unsigned int addr, unsigned int data)
{
  if (USE_DMA) ejtag_dma_write_h(addr, data);
  else ejtag_pracc_write_h2(addr, data);
}


static unsigned int ejtag_dma_read(unsigned int addr)
{
  unsigned int data;
  int retries = RETRY_ATTEMPTS;
  int retry2 = MAX_ATTEMPTS;

begin_ejtag_dma_read:

  // Setup Address
  set_instr(INSTR_ADDRESS);
  WriteData(addr);

  // Initiate DMA Read & set DSTRT
  set_instr(INSTR_CONTROL);
  ctrl_reg = ReadWriteData(DMAACC | DRWN | DMA_WORD | DSTRT | PROBEN | PRACC);

  // Wait for DSTRT to Clear
  while (ReadWriteData(DMAACC | PROBEN | PRACC) & DSTRT) 
    {if (retry2--)  
    	{test_reset();
    	 goto begin_ejtag_dma_read;}
      else 
      	 {printf("DMA Read Addr = %08x Set ADDR ERROR \n", addr); 
      	 	return(0xFFFF);}
    }

  retry2 = MAX_ATTEMPTS;
  // Read Data
  set_instr(INSTR_DATA);
  data = ReadData();

  // Clear DMA & Check DERR
  set_instr(INSTR_CONTROL);
  if (ReadWriteData(PROBEN | PRACC) & DERR)
    {
      if (retries--)  
      	{test_reset();
         goto begin_ejtag_dma_read;}
      else  printf("DMA Read Addr = %08x  Data = (%08x)ERROR ON READ\n", addr, data);
    }

  return(data);
}


static unsigned int ejtag_dma_read_h(unsigned int addr)
{
  unsigned int data;
  int retries = RETRY_ATTEMPTS;
  int retry2 = MAX_ATTEMPTS;

begin_ejtag_dma_read_h:

  // Setup Address
  set_instr(INSTR_ADDRESS);
  WriteData(addr);

  // Initiate DMA Read & set DSTRT
  set_instr(INSTR_CONTROL);
  ctrl_reg = ReadWriteData(DMAACC | DRWN | DMA_HALFWORD | DSTRT | PROBEN | PRACC);

  // Wait for DSTRT to Clear
  while (ReadWriteData(DMAACC | PROBEN | PRACC) & DSTRT)
   {if (retry2--)  
   	{test_reset();
     goto begin_ejtag_dma_read_h;}
    else 
      {printf("DMA Read(h) Addr = %08x  Set ADDR ERROR\n", addr); 
       	return(0xFFFF);}
    }
    retry2 = MAX_ATTEMPTS;
  // Read Data
  set_instr(INSTR_DATA);
  data = ReadData();

  // Clear DMA & Check DERR
  set_instr(INSTR_CONTROL);
  if (ReadWriteData(PROBEN | PRACC) & DERR)
    {
      if (retries--)  
      	{test_reset();
         goto begin_ejtag_dma_read_h;}
      else  printf("DMA Read(h) Addr = %08x  Data = (%08x)ERROR ON READ\n", addr, data);
    }

  // Handle half word endian mask
	if (BigEndian) // big
  	{if ( addr & 0x2 )  data = (data&0x0000ffff) ; //low
  		        else      data = (data>>16)&0x0000ffff ;  //high
		}	 
	else	//little
		{if ( addr & 0x2 )  data = (data>>16)&0x0000ffff ;   //high
  		        else      data = (data&0x0000ffff) ;   //low
		}
  return(data);
}


void ejtag_dma_write(unsigned int addr, unsigned int data)
{
  int   retries = RETRY_ATTEMPTS;
  int retry2 = MAX_ATTEMPTS;

begin_ejtag_dma_write:

  // Setup Address
  set_instr(INSTR_ADDRESS);
  WriteData(addr);

  // Setup Data
  set_instr(INSTR_DATA);
  WriteData(data);

  // Initiate DMA Write & set DSTRT
  set_instr(INSTR_CONTROL);
  ctrl_reg = ReadWriteData(DMAACC | DMA_WORD | DSTRT | PROBEN | PRACC);

  // Wait for DSTRT to Clear
  while (ReadWriteData(DMAACC | PROBEN | PRACC) & DSTRT)
     {if (retry2--)  
     	{test_reset();
       goto begin_ejtag_dma_write;}
     else 
     	 {printf("DMA Write Addr = %08x Set ADDR ERROR \n", addr); 
       	return;}
      }
    retry2 = MAX_ATTEMPTS;
  // Clear DMA & Check DERR
  set_instr(INSTR_CONTROL);
  if (ReadWriteData(PROBEN | PRACC) & DERR)
    {
      if (retries--)  
      	{test_reset();
         goto begin_ejtag_dma_write;}
      else  printf("DMA Write Addr = %08x  Data = ERROR ON WRITE\n", addr);
    }
}


void ejtag_dma_write_h(unsigned int addr, unsigned int data)
{
  int   retries = RETRY_ATTEMPTS;
  int retry2 = MAX_ATTEMPTS;

begin_ejtag_dma_write_h:

  // Setup Address
  set_instr(INSTR_ADDRESS);
  WriteData(addr);

  // Setup Data
  set_instr(INSTR_DATA);
  WriteData(data);

  // Initiate DMA Write & set DSTRT
  set_instr(INSTR_CONTROL);
  ctrl_reg = ReadWriteData(DMAACC | DMA_HALFWORD | DSTRT | PROBEN | PRACC);

  // Wait for DSTRT to Clear
  while (ReadWriteData(DMAACC | PROBEN | PRACC) & DSTRT)
      {if (retry2--)  
      	{test_reset();
         goto begin_ejtag_dma_write_h;}
       else 
      	 {printf("DMA Write(h) Addr = %08x Set ADDR ERROR \n", addr); 
      	 	return;}
      }
  retry2 = MAX_ATTEMPTS;
  // Clear DMA & Check DERR
  set_instr(INSTR_CONTROL);
  if (ReadWriteData(PROBEN | PRACC) & DERR)
    {
      if (retries--)  
      	{test_reset();
         goto begin_ejtag_dma_write_h;}
      else  printf("DMA Write(h) Addr = %08x  Data = ERROR ON WRITE\n", addr);
    }
}


static unsigned int ejtag_pracc_read(unsigned int addr)
{
  address_register = addr | 0xA0000000;  // Force to use uncached segment: kseg1
  data_register    = 0x0;
  ExecuteDebugModule(pracc_readword_code_module);
  return(data_register);
}

void ejtag_pracc_write(unsigned int addr, unsigned int data)
{
  address_register = addr | 0xA0000000;  // Force to use uncached segment: kseg1
  data_register    = data;
  ExecuteDebugModule(pracc_writeword_code_module);
}


static unsigned int ejtag_pracc_read_h(unsigned int addr)
{
  address_register = addr | 0xA0000000;  // Force to use uncached segment: kseg1
  data_register    = 0x0;
  ExecuteDebugModule(pracc_readhalf_code_module);
  return(data_register);
}


void ejtag_pracc_write_h(unsigned int addr, unsigned int data)
{
  address_register = addr | 0xA0000000;  // Force to use uncached segment: kseg1
  data_register    = data;
  ExecuteDebugModule(pracc_writehalf_code_module);
}
 // =================================================================================
 static unsigned int ejtag_pracc_read2(unsigned int addr)
{
  data_register    = 0x0;
  ExecuteDebugModule(ejtag_fix_readcode(addr, MIPS_WORD));
  return(data_register);
}

static unsigned int ejtag_pracc_read_h2(unsigned int addr)
{
  data_register    = 0x0;
  ExecuteDebugModule(ejtag_fix_readcode(addr, MIPS_HALFWORD));
  return(data_register);
}

void ejtag_pracc_write2(unsigned int addr, unsigned int data)
{
  data_register    = data;
  ExecuteDebugModule(ejtag_fix_writecode(addr, data, MIPS_WORD));
}

void ejtag_pracc_write_h2(unsigned int addr, unsigned int data)
{
  data_register    = data;
  ExecuteDebugModule(ejtag_fix_writecode(addr, data, MIPS_HALFWORD));
}

static unsigned int *ejtag_fix_readcode( unsigned int addr, int mode )
{
   unsigned short addr_hi, addr_lo;
   unsigned int *p;
	
   addr_lo = addr & 0xFFFF;   // %lo()
   addr_hi = ((addr >> 16) + (addr_lo >> 15)) | 0xA000;  //%hi(), and use uncached kseg1
     tnano(100);
   switch (mode)
   {
     case MIPS_WORD:
           p      =  pracc_readword_new;      
         *(p    ) = (*(p  ) & 0xFFFF0000) | (addr_hi & 0xFFFF);  //fill addr_hi
         *(p + 1) = (*(p+1) & 0xFFFF0000) | (addr_lo & 0xFFFC);  //fill addr_lo and align to word
          break;
                
                
     case MIPS_HALFWORD:
     	     p      =  pracc_readhalf_new;      
         *(p    ) = (*(p  ) & 0xFFFF0000) | (addr_hi & 0xFFFF);  //fill addr_hi
         *(p + 1) = (*(p+1) & 0xFFFF0000) | (addr_lo & 0xFFFE);  //fill addr_lo and align to halfword
          break;
                
     default:  
              printf("\n***ERROR: unsupported read data type!!! exit...\n");
              exit(1);
              break;
   }
	return p;
}

static unsigned int *ejtag_fix_writecode( unsigned int addr, unsigned int data, int mode )
{
   unsigned short addr_hi, addr_lo;
	 unsigned short data_hi, data_lo;
	 unsigned int *p;
	
   addr_lo = addr & 0xFFFF;   // %lo()
   addr_hi = ((addr >> 16) + (addr_lo >> 15)) | 0xA000;  //%hi(), and use uncached kseg1
      
   switch (mode)
   {
       case MIPS_WORD:
       	 data_hi = data >> 16;
         data_lo = data & 0xFFFF;
     	     p      =  pracc_writeword_new;      
         *(p    ) = (*(p  ) & 0xFFFF0000) | addr_hi;             //fill addr_hi
         *(p + 3) = (*(p+3) & 0xFFFF0000) | (addr_lo & 0xFFFC);  //fill addr_lo and align to word
         *(p + 1) = (*(p+1) & 0xFFFF0000) | data_hi;             //fill data_hi
         *(p + 2) = (*(p+2) & 0xFFFF0000) | data_lo;             //fill data_lo
          break;
          
     case MIPS_HALFWORD:
     	     data_lo = data & 0xFFFF;
           p      =  pracc_writehalf_new;      
         *(p    ) = (*(p  ) & 0xFFFF0000) | addr_hi;             //fill addr_hi
         *(p + 2) = (*(p+2) & 0xFFFF0000) | (addr_lo & 0xFFFE);  //fill addr_lo and align to halfword
         *(p + 1) = (*(p+1) & 0xFFFF0000) | data_lo;             //fill data_hi
          break;
                
                
   
                
     default:  
              printf("\n***ERROR: unsupported write data type!!! exit...\n");
              exit(1);
              break;
   }
	return p;
}
//  ===========================================================================

void ExecuteDebugModule(unsigned int *pmodule)
{
  unsigned int ctrl_reg;
  unsigned int address;
  unsigned int data   = 0;
  unsigned int offset = 0;
  int finished = 0;

   int   retries = RETRY_ATTEMPTS;

  if (DEBUGMSG) printf("DEBUGMODULE: Start module.\n");

  // Feed the chip an array of 32 bit values into the processor via the EJTAG port as instructions.
  while (1)
    {
      // Read the control register.  Make sure an access is requested, then do it.
      while (1)
        {
          set_instr(INSTR_CONTROL);
          ctrl_reg = ReadWriteData(PRACC | PROBEN | SETDEV);
          if (ctrl_reg & PRACC)
            break;
          if (DEBUGMSG) printf("DEBUGMODULE: No memory access in progress!\n");
        }

      set_instr(INSTR_ADDRESS);
      address = ReadData();

      // Check for read or write
      if (ctrl_reg & PRNW) // Bit set for a CPU WRITE to us
        {
          // Read the data out of CPU
          set_instr(INSTR_DATA);
          data = ReadData();
          
          // Clear the access pending bit (let the processor eat!)
          set_instr(INSTR_CONTROL);
          ctrl_reg = ReadWriteData(PROBEN | SETDEV);

          // Processor is writing to us
          if (DEBUGMSG1) printf("DEBUGMODULE: Write 0x%08X to address 0x%08X\n", data, address);
          // Handle Debug Write
          // If processor is writing to one of our psuedo virtual registers then save off data
          if (address == MIPS_VIRTUAL_ADDRESS_ACCESS)  address_register = data;
          if (address == MIPS_VIRTUAL_DATA_ACCESS)     data_register    = data;
        } //end cpu write

      else

        {
          // Check to see if its reading at the debug vector.  The first pass through
          // the module is always read at the vector, so the first one we allow.  When
          // the second read from the vector occurs we are done and just exit.
          if (address == MIPS_DEBUG_VECTOR_ADDRESS)
            {
              if (finished++) // Allows ONE pass ,we use a "b" instuction at the end, gp will jump back to start
                {
                  if (DEBUGMSG1) printf("DEBUGMODULE: Finished module.\n");
                  return;
                }
            }

          // Processor is reading from us
          if (address >= MIPS_DEBUG_VECTOR_ADDRESS)
            {
              // Reading an instruction from our module so fetch the instruction from the module
              offset = (address - MIPS_DEBUG_VECTOR_ADDRESS) / 4;
              data = *(unsigned int *)(pmodule + offset);    //debug instruction, same for both endians. No reverse required for BIG!
              if (DEBUGMSG1) printf("DEBUGMODULE: Instruction read at 0x%08X  offset -> %04d  data -> 0x%08X\n", address, offset, data); //fflush(stdout);
            }
          else
            {
              // Reading from our virtual register area
              if (DEBUGMSG1) printf("DEBUGMODULE: Read address 0x%08X  data = 0x%08X\n", address, data);
              // Handle Debug Read
              // If processor is reading from one of our psuedo virtual registers then give it data
              if (address == MIPS_VIRTUAL_ADDRESS_ACCESS)  data = address_register;
              if (address == MIPS_VIRTUAL_DATA_ACCESS)     data = data_register;
            }

          // Send the data out
          set_instr(INSTR_DATA);
          data = ReadWriteData(data);

          // Clear the access pending bit (let the processor eat!)
          set_instr(INSTR_CONTROL);
          ctrl_reg = ReadWriteData(PROBEN | SETDEV);

        } //end cpu read
    }
}


void chip_detect(void)
{
  unsigned int id = 0x0;

  processor_chip_type*   processor_chip = processor_chip_list;

  lpt_openport();

  printf("Probing bus ... ");

  if (skipdetect)
    {
      // Manual Override CPU Chip ID
      test_reset();
      instruction_length = instrlen;
      endian = (endian + force_rev_endian) & 0x1 ;  //reverse endian
      if(BigEndian)   //big endian
      	{ 
      		printf("CPU running under BIG endian\n\n");
      	}
      else
      	printf("CPU running under LITTLE endian\n\n");
      set_instr(INSTR_IDCODE);
      id = ReadData();
      printf("Done\n\n");
      printf("Instruction Length set to %d\n\n",instruction_length);
      printf("CPU Chip ID: ");
      ShowData(id);

      printf("*** CHIP DETECTION OVERRIDDEN ***\n\n");
      return;
    }
  else
    {
      // Auto Detect CPU Chip ID
      while (processor_chip->chip_id)
        {
          test_reset();
          if (instrlen)
            instruction_length = instrlen;
          else
            instruction_length = processor_chip->instr_length;

          endian = (processor_chip->chip_endn + force_rev_endian) & 0x1 ;  //reverse endian

          set_instr(INSTR_IDCODE);
          id = ReadData();
          if (id == processor_chip->chip_id)
            {
              init_code = processor_chip->init_code;
              printf("Done\n\n");
              printf("Instruction Length set to %d\n\n",instruction_length);
              if (BigEndian) printf("CPU running under BIG endian\n\n");
              	else printf("CPU running under LITTLE endian\n\n");
              printf("CPU Chip ID: ");
              ShowData(id);
              printf("*** Found a %s chip ***\n\n", processor_chip->chip_descr);
              return;
            }
          processor_chip++;
        }
    }

  printf("Done\n\n");
  printf("Instruction Length set to %d\n\n",instruction_length);
  printf("CPU Chip ID: ");
  ShowData(id);
  printf("*** Unknown or NO CPU Chip ID Detected ***\n\n");

  printf("*** Possible Causes:\n");
  printf("    1) Router/Modem is not Connected.\n");
  printf("    2) Router/Modem is not Powered On.\n");
  printf("    3) Improper JTAG Cable.\n");
  printf("    4) Unrecognized CPU Chip ID.\n");

  chip_shutdown();;
  exit(0);
}


void check_ejtag_features()
{
  unsigned int features;

  set_instr(INSTR_IMPCODE);
  features = ReadData();

  printf("    - EJTAG IMPCODE ....... : ");
  ShowData(features);

  // EJTAG Version
  ejtag_version = (features >> 29) & 7;
  printf("    - EJTAG Version ....... : ");
  if (ejtag_version == 0)       printf("1 or 2.0\n");
  else if (ejtag_version == 1)  printf("2.5\n");
  else if (ejtag_version == 2)  printf("2.6\n");
  else if (ejtag_version == 3)  printf("3.1\n");
  else                          printf("Unknown (%d is a reserved value)\n", ejtag_version);

  // EJTAG DMA Support
  USE_DMA = !(features & (1 << 14));
  printf("    - EJTAG DMA Support ... : %s\n", USE_DMA ? "Yes" : "No");
  printf( "    - EJTAG Implementation flags:%s%s%s%s%s%s%s\n",
          (features & (1 << 28)) ? " R3k"	: " R4k",
          (features & (1 << 24)) ? " DINTsup"	: "",
          (features & (1 << 22)) ? " ASID_8"	: "",
          (features & (1 << 21)) ? " ASID_6"	: "",
          (features & (1 << 16)) ? " MIPS16"	: "",
          (features & (1 << 14)) ? " NoDMA"	: "",
          (features & (1      )) ? " MIPS64"	: " MIPS32" );

  if (force_dma)
    {
      USE_DMA = 1;
      printf("    *** DMA Mode Forced On ***\n");
    }
  if (force_nodma)
    {
      USE_DMA = 0;
      printf("    *** DMA Mode Forced Off ***\n");
    }

  printf("\n");
}


void chip_shutdown(void)
{
  fflush(stdout);
  test_reset();
  lpt_closeport();
}

/**** sp_xxxxxx() is a function specific to Spansion flash ****/
void sp_unlock_bypass(void)
{
  sp_exit_cmdset();    //exit any command set
  ejtag_write_h(FLASH_MEMORY_START + (0x555 << 1), 0x00aa00aa ); /* unlock bypass */
  ejtag_write_h(FLASH_MEMORY_START + (0x2aa << 1), 0x00550055 );
  ejtag_write_h(FLASH_MEMORY_START + (0x555 << 1), 0x00200020 );
  printf("\nEntered Unlock Bypass mode->\n");
}

void sp_unlock_bypass_reset(void)
{
  sp_exit_cmdset();
}

void sp_exit_cmdset(void)
{
  ejtag_write_h(FLASH_MEMORY_START + (0x555 << 1), 0x00900090 ); /* exit command set */
  ejtag_write_h(FLASH_MEMORY_START + (0x555 << 1), 0x00000000 );
}

void sp_exit_cmdset2(void)
{
  ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00900090 ); /* exit command set */
  ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00000000 );
}

/***  Check Flash Protection Status   ***/
void sp_check_ppb(void)
{
	unsigned int glbppb=0, lockreg=0, sppb=0, cur_block, block_addr;
  unsigned int tmp;
	
	sflash_reset();
	sp_reset();
	printf("\n\nFlash Sector Protection type %1d\n\n",PPB_type);
	
  if (PPB_type==SP_PPB)
  {
    if (!issue_chip_erase && (issue_show_ppb || issue_clear_ppb))
 	  { 
       // read Lock Register
       sp_exit_cmdset2();    //exit any command set
       ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA ); /* Lock Register Command Set Entry */
       ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055 );
       ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00400040 );
       lockreg = ejtag_read_h(FLASH_PROBE_BASE );
       sp_exit_cmdset2();    //exit any command set
       printf("\nRead Lock Register Status: ");
              ShowData_h(lockreg);
 
       // read Global PPB Lock status
        sp_exit_cmdset2();    //exit any command set
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA ); /* Global PPB Lock Command Set Entry */
        ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055 );
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00500050 );
        glbppb = ejtag_read_h(FLASH_PROBE_BASE);
        sp_exit_cmdset2();    //exit any command set
        printf("\nRead Globle PPB Lock Status: ");
               ShowData_h(glbppb);
                    
       // read Non-volatile Sector PPBs
        for (cur_block = 1;  cur_block <= block_total;  cur_block++)
         {
            block_addr = blocks[cur_block];
            sp_exit_cmdset2();    //exit any command set
            ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA ); /* Sector PPB Command Set Entry */
            ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055 );
            ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00C000C0 );
            sppb = ejtag_read_h(block_addr);
            sp_exit_cmdset2();    //exit any command set
          
            printf("Read Sector: %d (addr = %08x) with PPB:", cur_block, block_addr);
            fflush(stdout);
               ShowData_h(sppb);
            fflush(stdout);
            sflash_reset();
	          sp_reset();
         }
      }
    //clear PPB to unprotect 
    if (!issue_chip_erase && issue_clear_ppb)
	    {
        printf("Erase all Sector PPBs...");   
        sp_exit_cmdset2();    //exit any command set
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA ); // Sector PPB Command Set Entry 
        ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055 );
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00C000C0 );
        ejtag_write_h(FLASH_PROBE_BASE , 0x00800080 ); // Erase all sector PPB 
        ejtag_write_h(FLASH_PROBE_BASE , 0x00300030 );

        while ( (tmp=ejtag_read_h(FLASH_PROBE_BASE)) != 0x1 ) 
         { if (DEBUGMSG1) printf("%08X, poll on PPB erase!...\n",tmp);}
        mysleep(2);
        sp_exit_cmdset2();    //exit any command set            
        printf("   Done!\n\n"); 
    
     } 
    
   if (issue_chip_erase)    // Erase all non-volatile PPB
     {	
   	    sflash_reset();
   	    sp_reset();
   
   			printf("Erasing will take 1 or 2 minutes,Don't break!!!\nErase whole chip ...");
   			sp_exit_cmdset2();    //exit any command set
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA ); /* Unlock*/
        ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055 );
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00800080 );
              //Erase Chip
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA);
        ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055);
        ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00100010);

        sflash_poll(FLASH_PROBE_BASE, 0xFFFF);
        printf("   Done!\n\n");
        sflash_reset();
	      sp_reset();
	  
     }  //end PPB process
  }   //end Spansion PPB protectio
}

void sp_reset(void)
{
  ejtag_write_h(FLASH_PROBE_BASE, 0x00F000F0);    // Set array to read mode
}

void sp_init_cpu()
{
  address_register = 0xBFC000000;  // just here,not used
  data_register    = 0x0;          // just here,not used
  ExecuteDebugModule(init_code);
}

// initialize $1 register with 0xFF200000, PraCC vitual data register base address.
void sp_init_dreg()
{
  address_register = 0xBFC000000;  // just here,not used
  data_register    = 0x0;          // just here,not used
  ExecuteDebugModule(pracc_init_dreg);
}

void sp_get_chipid(void)
{
  /*     Spansion ID workaround (kb1klk) 30 Dec 2007

       Vendor ID altered to 017E to avoid ambiguity in lookups.
       Spansion uses vend 0001 dev 227E for numerous chips.
       This code reads device SubID from its address and combine this into
       the unique devid that will be used in the lookup table.                */

  unsigned int devsubid_m, devsubid_l;
  vendid = 0x017E;
  devsubid_m = 0x00FF & ejtag_read_h(FLASH_PROBE_BASE+(0x0E << 1));  // Vendor ID cycle 2
  devsubid_l = 0x00FF & ejtag_read_h(FLASH_PROBE_BASE+(0x0F << 1));  // Vendor ID cycle 3
  devid = (devsubid_m << 8) | devsubid_l;

}

unsigned long rev_endian(unsigned long data)
{
// swap whole word  
    return ((data << 24) & 0xff000000) |
           ((data <<  8) & 0x00ff0000) |
           ((data >>  8) & 0x0000ff00) |
           ((data >> 24) & 0x000000ff) ;
}

unsigned long rev_endian_h(unsigned long data)
{
// swap lower bytes  
    return  ((data << 8) & 0x0000ff00) |
            ((data >> 8) & 0x000000ff) ;
}


void run_backup(char *filename, unsigned int start, unsigned int length)
{
  unsigned int addr, data;
  FILE *fd;
  int counter = 0;
  int percent_complete = 0;
  char newfilename[128] = "";
  time_t start_time = time(0);
  time_t end_time, elapsed_seconds;
  int align=0;
  unsigned char *overead;
  

  struct tm* lt = localtime(&start_time);
  char time_str[15];

 
  sprintf(time_str, "%04d%02d%02d_%02d%02d%02d",
          lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
          lt->tm_hour, lt->tm_min, lt->tm_sec
         );

  printf("*** You Selected to Backup the %s ***\n\n",filename);

  strcpy(newfilename,filename);
  strcat(newfilename,".SAVED");
  if (issue_timestamp)
    {
      strcat(newfilename,"_");
      strcat(newfilename,time_str);
    }

  fd = fopen(newfilename, "wb" );
  if (fd<=0)
    {
      fprintf(stderr,"Could not open %s for writing\n", newfilename);
      exit(1);
    }

  printf("=========================\n");
  printf("Backup Routine Started\n");
  printf("=========================\n");

  printf("\nSaving %s to Disk...\n",newfilename);
  
  for (addr=start; addr<(start+length); addr+=4)
    {
      counter += 4;
      percent_complete = (counter * 100 / length);
      if (!silent_mode)
        if ((addr&0xF) == 0)  printf("[%3d%% Backed Up]   %08x: ", percent_complete, addr);
      data = ejtag_read(addr);  //whole word(32bits) memory operation
      if(BigEndian) data= rev_endian(data);
         fwrite( (unsigned char*) &data, 1, sizeof(data), fd);

      if (silent_mode)  printf("%4d%%   bytes = %d\r", percent_complete, counter);
      else              printf("%08x%c", data, (addr&0xF)==0xC?'\n':' ');

      fflush(stdout);
    }
  fclose(fd);

  printf("Done  (%s saved to Disk OK)\n\n",newfilename);

  printf("bytes written: %d\n", counter);

  printf("=========================\n");
  printf("Backup Routine Complete\n");
  printf("=========================\n");

  time(&end_time);
  elapsed_seconds = difftime(end_time, start_time);
  printf("elapsed time: %d seconds\n", (int)elapsed_seconds);
}



void run_flash(char *filename, unsigned int start, unsigned int length)
{
  unsigned int addr, data ;
  FILE *fd ;
  int counter = 0;
  int percent_complete = 0;
  time_t start_time = time(0);
  time_t end_time, elapsed_seconds;

  printf("*** You Selected to Flash the %s ***\n\n",filename);

  fd=fopen(filename, "rb" );
  if (fd<=0)
    {
      fprintf(stderr,"Could not open %s for reading\n", filename);
      exit(1);
    }

  printf("=========================\n");
  printf("Flashing Routine Started\n");
  printf("=========================\n");

  if (issue_erase) sflash_erase_area(start,length);

  if (bypass)
    {
      sp_unlock_bypass();
    }

  printf("\nLoading %s to Flash Memory...\n",filename);
  for (addr=start; addr<(start+length); addr+=4)
    {
      counter += 4;
      percent_complete = (counter * 100 / length);
      if (!silent_mode)
        if ((addr&0xF) == 0)  printf("[%3d%% Flashed]   %08x: ", percent_complete, addr);

      fread( (unsigned char*) &data, 1,sizeof(data), fd);
      if (BigEndian) data = rev_endian(data);
      // Erasing Flash Sets addresses to 0xFF's so we can avoid writing these (for speed)
      if (issue_erase)
        {
          if (!(data == 0xFFFFFFFF))
            sflash_write_word(addr, data);  //send a word 32bits to write
        }
      else sflash_write_word(addr, data);  // Otherwise we gotta flash it all

      if (silent_mode)  printf("%4d%%   bytes = %d\r", percent_complete, counter);
      else              printf("%08x%c", data, (addr&0xF)==0xC?'\n':' ');


      fflush(stdout);
      data = 0xFFFFFFFF;  // This is in case file is shorter than expected length
    }
  fclose(fd);
  printf("Done  (%s loaded into Flash Memory OK)\n\n",filename);

  sflash_reset();

  printf("=========================\n");
  printf("Flashing Routine Complete\n");
  printf("=========================\n");

  time(&end_time);
  elapsed_seconds = difftime(end_time, start_time);
  printf("elapsed time: %d seconds\n", (int)elapsed_seconds);
}


void run_erase(char *filename, unsigned int start, unsigned int length)
{
  time_t start_time = time(0);
  time_t end_time, elapsed_seconds;

  printf("*** You Selected to Erase the %s ***\n\n",filename);

  printf("=========================\n");
  printf("Erasing Routine Started\n");
  printf("=========================\n");

  sflash_erase_area(start,length);
  sflash_reset();

  printf("=========================\n");
  printf("Erasing Routine Complete\n");
  printf("=========================\n");

  time(&end_time);
  elapsed_seconds = difftime(end_time, start_time);
  printf("elapsed time: %d seconds\n", (int)elapsed_seconds);
}


void identify_flash_part(void)
{
  flash_chip_type*   flash_chip = flash_chip_list;
  flash_area_type*   flash_area = flash_area_list;

  unsigned int topblocksize=0;

  // Important for these to initialize to zero
  block_addr  = 0;
  block_total = 0;
  flash_size  = 0;
  cmd_type    = 0;
  PPB_type    = 0;
  strcpy(flash_part,"");
  

  while (flash_chip->vendid)
    {
      if ((flash_chip->vendid == vendid) && (flash_chip->devid == devid))
        {
          flash_size = flash_chip->flash_size;
          cmd_type   = flash_chip->cmd_type;
          PPB_type   = flash_chip->prot_type;
          strcpy(flash_part, flash_chip->flash_part);
          
          if (flash_chip->region5_num) topblocksize = flash_chip->region5_size;
          	else if (flash_chip->region4_num) topblocksize = flash_chip->region4_size;
          		else if (flash_chip->region3_num) topblocksize = flash_chip->region3_size;
          			else if (flash_chip->region2_num) topblocksize = flash_chip->region2_size;
          				else if (flash_chip->region1_num) topblocksize = flash_chip->region1_size;
          topblocksize =(topblocksize >= NVRAM_LEN)? topblocksize : NVRAM_LEN;  // get flash top reserved space
          	
			if (BigEndian) 		// IF flash size>4MB, re-calculate address and length for ADSL chip
					{
					if (flash_size > size4MB) FLASH_MEMORY_START = MEM_TOP - flash_size;
            else FLASH_MEMORY_START = 0x1FC00000;
          } 
       else     //little endian 
          {  
          if (flash_size >= size8MB) FLASH_MEMORY_START = 0x1C000000;   //keep old setting for Wifi chip
          else FLASH_MEMORY_START = 0x1FC00000;
          }	

          while (flash_area->chip_size)
            {
              if ((flash_area->chip_size == flash_size) && (strcasecmp(flash_area->area_name, AREA_NAME)==0))
                {
                 AREA_START  = flash_area->area_start;
                 AREA_LENGTH = flash_area->area_length;
                 
                 if(BigEndian &&(flash_size > size4MB) )           //BCM63x8, 8MB & 16MB
                 	 {
                   if (strcasecmp(AREA_NAME,"TFE")==0)              //one bottom block size
                      { AREA_START  = FLASH_MEMORY_START;
                        AREA_LENGTH = flash_chip->region1_size; }
                        
                   else if (strcasecmp(AREA_NAME,"WHOLEFLASH")==0)
                      { AREA_START  = FLASH_MEMORY_START;
                        AREA_LENGTH = flash_size; }
                        
                   else if (strcasecmp(AREA_NAME,"TERNEL")==0)   
                      { AREA_START  = FLASH_MEMORY_START + flash_chip->region1_size;
                        AREA_LENGTH = size4MB - flash_chip->region1_size; }
                        
                   else if (strcasecmp(AREA_NAME,"NVRAM")==0)
                      { AREA_START  = MEM_TOP - topblocksize * 2;
                        AREA_LENGTH = topblocksize * 2; }
                        
                   else if (strcasecmp(AREA_NAME,"CFE")==0)
                      { AREA_START  = FLASH_MEMORY_START;}
                      
                   else if (strcasecmp(AREA_NAME,"KERNEL")==0)
                      { AREA_START  = FLASH_MEMORY_START + CFE_LEN;}    
                 	 }
                 	 
                 strcat(AREA_NAME,".BIN");
                 break;
                }
              flash_area++;
            }

          if (strcasecmp(AREA_NAME,"CUSTOM")==0)
            {
              strcat(AREA_NAME,".BIN");
              FLASH_MEMORY_START = selected_window;
              AREA_START         = selected_start;
              AREA_LENGTH        = selected_length;
              if ((AREA_START < FLASH_MEMORY_START) || 
              	  (AREA_START >= FLASH_MEMORY_START + flash_size  )  ||
              	  (AREA_START + AREA_LENGTH <= FLASH_MEMORY_START )  ||
              	  (AREA_START + AREA_LENGTH > FLASH_MEMORY_START + flash_size))  
                printf("\n\nWarning: Custom area NOT in the FLASH memory window!!!\n\n");  
              
            }

          if (flash_chip->region1_num)  define_block(flash_chip->region1_num, flash_chip->region1_size);
          if (flash_chip->region2_num)  define_block(flash_chip->region2_num, flash_chip->region2_size);
          if (flash_chip->region3_num)  define_block(flash_chip->region3_num, flash_chip->region3_size);
          if (flash_chip->region4_num)  define_block(flash_chip->region4_num, flash_chip->region4_size);
          if (flash_chip->region5_num)  define_block(flash_chip->region5_num, flash_chip->region5_size);

          sflash_reset();
      	  printf("Matching Flash Chip (VenID:DevID = %04x : %04x)\n\n",vendid,devid);
          
          if (selected_fc != 0)
            printf("*** Manually Selected a %s Flash Chip ***\n\n", flash_part);
          else
            printf("*** Found a %s Flash Chip ***\n\n", flash_part);

          printf("    - Flash Chip Window Start .... : %08x\n", FLASH_MEMORY_START);
          printf("    - Flash Chip Window Length ... : %08x\n", flash_size);
          printf("    - Selected Area Start ........ : %08x\n", AREA_START);
          printf("    - Selected Area Length ....... : %08x\n\n", AREA_LENGTH);

          break;
        }
      flash_chip++;
    }
}

// get start address of each block, max block number
void define_block(unsigned int block_count, unsigned int block_size)
{
  unsigned int  i;

  if (block_addr == 0)  block_addr = FLASH_MEMORY_START;

  for (i = 1; i <= block_count; i++)
    {
      block_total++;
      blocks[block_total] = block_addr;
      block_addr = block_addr + block_size;
    }
  blocks[block_total+1] = block_addr; //get end of flash address+1
}


void sflash_config(void)
{
  flash_chip_type*   flash_chip = flash_chip_list;
  int counter = 0;

  printf("\nManual Flash Selection ... ");

  while (flash_chip->vendid)
    {
      counter++;
      if (counter == selected_fc)
        {
          vendid = flash_chip->vendid;
          devid  = flash_chip->devid;
          identify_flash_part();
          break;
        }
      flash_chip++;
    }

  if (strcasecmp(flash_part,"")==0)
    printf("*** Unknown or NO Flash Chip Selected ***\n");

}

void sflash_probe(void)
{
  int retries = MAX_ATTEMPTS;

  // Use default flash window for flash probe if not customized the probe address.
  if (strcasecmp(AREA_NAME,"CUSTOM")==0)   FLASH_MEMORY_START = selected_window;	
      else        FLASH_MEMORY_START = 0x1FC00000;
  	     
  if (!probeme) FLASH_PROBE_BASE = FLASH_MEMORY_START;
   	
  printf("\nProbing Flash at (Probe Address: 0x%08x) ...\n", FLASH_PROBE_BASE);

again:

  strcpy(flash_part,"");

  // Probe using cmd_type for AMD
  if (strcasecmp(flash_part,"")==0)
    {

      cmd_type = CMD_TYPE_AMD;
      sflash_reset();
      ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_PROBE_BASE + (0x2AA << 1), 0x00550055);
      ejtag_write_h(FLASH_PROBE_BASE + (0x555 << 1), 0x00900090);
      vendid = ejtag_read_h(FLASH_PROBE_BASE);      //get manufacture ID
      devid  = ejtag_read_h(FLASH_PROBE_BASE+2);    //get Vendor ID cycle 1
      if (((vendid & 0xFF) == 0x0001) && (devid == 0x227E)) sp_get_chipid();  //get spansion 3 bytes VenID
      identify_flash_part();
      if (strcasecmp(flash_part,"")==0)
 	           printf("Detected Chip ID with AMD/Spansion Routine (VenID:DevID = %04x : %04x)\n",vendid,devid);

    }

  // Probe using cmd_type for SST
  if (strcasecmp(flash_part,"")==0)
    {
      cmd_type = CMD_TYPE_SST;
      sflash_reset();
      ejtag_write_h(FLASH_PROBE_BASE + (0x5555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_PROBE_BASE + (0x2AAA << 1), 0x00550055);
      ejtag_write_h(FLASH_PROBE_BASE + (0x5555 << 1), 0x00900090);
      vendid = ejtag_read_h(FLASH_PROBE_BASE);
      devid  = ejtag_read_h(FLASH_PROBE_BASE+2);
      identify_flash_part();
      if (strcasecmp(flash_part,"")==0)
 	           printf("Detected Chip ID with SST Routine (VenID:DevID = %04x : %04x)\n",vendid,devid);

    }

  // Probe using cmd_type for BSC & SCS
  if (strcasecmp(flash_part,"")==0)
    {
      cmd_type = CMD_TYPE_BSC;
      sflash_reset();
      ejtag_write_h(FLASH_PROBE_BASE, 0x00900090);
      vendid = ejtag_read_h(FLASH_PROBE_BASE);
      devid  = ejtag_read_h(FLASH_PROBE_BASE+2);
      identify_flash_part();
      if (strcasecmp(flash_part,"")==0)
 	            printf("Detected Chip ID with BSC/SCS Routine (VenID:DevID = %04x : %04x)\n",vendid,devid);

    }

  if (strcasecmp(flash_part,"")==0)
    {
      if (retries--)
        goto again;
      else
        {
          printf("Done\n\n");
          printf("*** Unknown or NO Flash Chip Detected ***\n");
         }
    }

  return;
}

void sflash_poll(unsigned int addr, unsigned int data)
{
  if ((cmd_type == CMD_TYPE_BSC) || (cmd_type == CMD_TYPE_SCS))
    {
      // Wait Until Ready
      while ( (ejtag_read_h(FLASH_MEMORY_START) & STATUS_READY) != STATUS_READY );
    }
  else
    {
      // Wait Until Ready
      while ( (ejtag_read_h(addr) & STATUS_READY) != (data & STATUS_READY) );
    }

}


void sflash_erase_area(unsigned int start, unsigned int length)
{
  int cur_block;
  int tot_blocks;
  unsigned int reg_start;
  unsigned int reg_end;         
  int block_start=0;
  int block_end=0;
  int align_start=0;
  int align_end=0;
  unsigned int block_addr2;
 
  reg_start = start;
  reg_end   = reg_start + length;   //the end +1

  tot_blocks = 0;

  for (cur_block = 1;  cur_block <= block_total;  cur_block++)
    {
      block_addr = blocks[cur_block];
      block_addr2 = blocks[cur_block+1];  //current block end +1
     if ((reg_start >= block_addr) && (reg_start < block_addr2)) block_start = cur_block;
     if ((block_start>0) && (reg_end >= block_addr) && (reg_end < block_addr2)) block_end = cur_block;	
     if (reg_start == block_addr) align_start = 1;
     if ((length>0) && (reg_end == block_addr2)) {align_end = 1; block_end = cur_block+1;}
    }
// printf("Erase start from sect %8d to sect %8d\n", block_start, block_end);   
    
  if (block_start == 0 || block_end == 0 )
  	{ printf("***ERROR: No blocks need to Erase or Window setting wrong!!\n");
  	goto go_out; }
  	
  if ((!force_align)&& ((!align_start) || (!align_end)))
  	{ printf("***ERROR: Erase or Program area doesn't align with flash block bound!\n"
  				   "          Modify oprating area or use /force_align option.\n");
  	goto go_out;	}
  	
  if(align_end) tot_blocks = block_end - block_start;
  	 else tot_blocks = block_end - block_start + 1;
  printf("Total Blocks to Erase: %d\n\n", tot_blocks);

  for (cur_block = block_start;  cur_block < block_start + tot_blocks;  cur_block++)
   {
     block_addr = blocks[cur_block];
     printf("Erasing block: %d (addr = %08x)...", cur_block, block_addr);
     fflush(stdout);
     sflash_erase_block(block_addr);
     printf("Done\n");
     fflush(stdout);
    }
        
go_out:
 return;
}


void sflash_erase_block(unsigned int addr)
{

  if (cmd_type == CMD_TYPE_AMD)
    {
      sflash_reset();
      //Unlock Block
      ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_MEMORY_START+(0x2AA << 1), 0x00550055);
      ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00800080);

      //Erase Block
      ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_MEMORY_START+(0x2AA << 1), 0x00550055);
      ejtag_write_h(addr, 0x00300030);


      // Wait for Erase Completion
      sflash_poll(addr, 0xFFFF);

    }

  if (cmd_type == CMD_TYPE_SST)
    {

      //Unlock Block
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_MEMORY_START+(0x2AAA << 1), 0x00550055);
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00800080);

      //Erase Block
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_MEMORY_START+(0x2AAA << 1), 0x00550055);
      ejtag_write_h(addr, 0x00500050);

      // Wait for Erase Completion
      sflash_poll(addr, 0xFFFF);

    }

  if ((cmd_type == CMD_TYPE_BSC) || (cmd_type == CMD_TYPE_SCS))
    {

      //Unlock Block
      ejtag_write_h(addr, 0x00500050);     // Clear Status Command
      ejtag_write_h(addr, 0x00600060);     // Unlock Flash Block Command
      ejtag_write_h(addr, 0x00D000D0);     // Confirm Command

      // Wait for Unlock Completion
      sflash_poll(addr, STATUS_READY);

      //Erase Block
      ejtag_write_h(addr, 0x00500050);     // Clear Status Command
      ejtag_write_h(addr, 0x00200020);     // Block Erase Command
      ejtag_write_h(addr, 0x00D000D0);     // Confirm Command

      // Wait for Erase Completion
      sflash_poll(addr, STATUS_READY);

    }

  sflash_reset();

}


void sflash_reset(void)
{

  if ((cmd_type == CMD_TYPE_AMD) || (cmd_type == CMD_TYPE_SST))
    {
      ejtag_write_h(FLASH_MEMORY_START, 0x00F000F0);    // Set array to read mode
    }

  if ((cmd_type == CMD_TYPE_BSC) || (cmd_type == CMD_TYPE_SCS))
    {
      ejtag_write_h(FLASH_MEMORY_START, 0x00500050);    // Clear CSR
      ejtag_write_h(FLASH_MEMORY_START, 0x00ff00ff);    // Set array to read mode
    }

}

void sflash_write_word(unsigned int addr, unsigned int data)
{
  unsigned int data_lo, data_hi, dbg;
  dbg = 1;

   if (USE_DMA)
    {
      // DMA Uses Byte Lanes
     data_lo = (data & 0x0000FFFF) ;
     data_hi = (data & 0xFFFF0000);   
    }
  else
    {
      // PrAcc Does Not
      data_lo = (data & 0xFFFF);
      data_hi = ((data >> 16) & 0xFFFF);
    } 
 
    

  if (cmd_type == CMD_TYPE_AMD)
    {
      if (bypass)
        {
          ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00A000A0);
          if (!BigEndian) ejtag_write_h(addr, data_lo);             //little endian
          	else ejtag_write_h(addr+2, data_lo);                    //big endian
          tnano(100);

          ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00A000A0);
          
          if (!BigEndian) ejtag_write_h(addr+2, data_hi);
          	else ejtag_write_h(addr, data_hi);
          
          tnano(100);
        }

      else
        {
          ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00AA00AA);
          ejtag_write_h(FLASH_MEMORY_START+(0x2AA << 1), 0x00550055);
          ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00A000A0);
//          ejtag_write_h(addr, data_lo);
          if (!BigEndian) ejtag_write_h(addr, data_lo);
          	else ejtag_write_h(addr+2, data_lo);

          // Wait for Completion
//          sflash_poll(addr, (data & 0xffff));
					if (!BigEndian) sflash_poll(addr, (data & 0xffff));
          	else sflash_poll(addr+2, (data & 0xffff));

          // Now Handle Other Half Of Word
          ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00AA00AA);
          ejtag_write_h(FLASH_MEMORY_START+(0x2AA << 1), 0x00550055);
          ejtag_write_h(FLASH_MEMORY_START+(0x555 << 1), 0x00A000A0);
//          ejtag_write_h(addr+2, data_hi);
          if (!BigEndian) ejtag_write_h(addr+2, data_hi);
          	else ejtag_write_h(addr, data_hi);

          // Wait for Completion
//          sflash_poll(addr+2, ((data >> 16) & 0xffff));
					if (!BigEndian) sflash_poll(addr+2, ((data >> 16) & 0xffff));
          	else sflash_poll(addr, ((data >> 16) & 0xffff));
        }

    }

  if (cmd_type == CMD_TYPE_SST)
    {
      // Handle Half Of Word
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_MEMORY_START+(0x2AAA << 1), 0x00550055);
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00A000A0);
//      ejtag_write_h(addr, data_lo);
          if (!BigEndian) ejtag_write_h(addr, data_lo);
          	else ejtag_write_h(addr+2, data_lo);

      // Wait for Completion
//      sflash_poll(addr, (data & 0xffff));
					if (!BigEndian) sflash_poll(addr, (data & 0xffff));
          	else sflash_poll(addr+2, (data & 0xffff));


      // Now Handle Other Half Of Word
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00AA00AA);
      ejtag_write_h(FLASH_MEMORY_START+(0x2AAA << 1), 0x00550055);
      ejtag_write_h(FLASH_MEMORY_START+(0x5555 << 1), 0x00A000A0);
//      ejtag_write_h(addr+2, data_hi);
          if (!BigEndian) ejtag_write_h(addr+2, data_hi);
          	else ejtag_write_h(addr, data_hi);


      // Wait for Completion
//      sflash_poll(addr+2, ((data >> 16) & 0xffff));
					if (!BigEndian) sflash_poll(addr+2, ((data >> 16) & 0xffff));
          	else sflash_poll(addr, ((data >> 16) & 0xffff));

    }

  if ((cmd_type == CMD_TYPE_BSC) || (cmd_type == CMD_TYPE_SCS))
    {
      // Handle Half Of Word
      ejtag_write_h(addr, 0x00500050);     // Clear Status Command
      ejtag_write_h(addr, 0x00400040);     // Write Command
//      ejtag_write_h(addr, data_lo);        // Send HalfWord Data
          if (!BigEndian) ejtag_write_h(addr, data_lo);
          	else ejtag_write_h(addr, data_hi);

      ejtag_write_h(addr, 0x00700070);     // Check Status Command

      // Wait for Completion
      sflash_poll(addr, STATUS_READY);


      // Now Handle Other Half Of Word
      ejtag_write_h(addr+2, 0x00500050);   // Clear Status Command
      ejtag_write_h(addr+2, 0x00400040);   // Write Command
//      ejtag_write_h(addr+2, data_hi);      // Send HalfWord Data
          if (!BigEndian) ejtag_write_h(addr+2, data_hi);
          	else ejtag_write_h(addr+2, data_lo);

      ejtag_write_h(addr+2, 0x00700070);   // Check Status Command

      // Wait for Completion
      sflash_poll(addr+2, STATUS_READY);
    }
}


void show_usage(void)
{

  flash_chip_type*      flash_chip = flash_chip_list;
  processor_chip_type*  processor_chip = processor_chip_list;
  int counter = 0;

  printf( " ABOUT: This program reads/writes flash memory on the Broadcom MIPS(LE)\n"
          "        Chip and compatible routers via EJTAG using either DMA Access\n"
          "        routines or PrAcc routines (slower/more compatible). Processor chips\n"
          "        supported in this version include the following chips:\n\n"
          "            Supported Chips\n"
          "            ---------------\n");

  while (processor_chip->chip_id)
    {
      printf("            %-40.40s\n", processor_chip->chip_descr);
      processor_chip++;
    }


  printf( "\n\n");
  printf( " USAGE: BrJtag [parameter] </noreset> </noemw> </nocwd> </nobreak></revendian>\n"
          "                      </notimestamp> </dma> </nodma> </noerase></initcpu>\n"
          "                      </bypass></forcealign></showppb></clearppb></erasechip>\n"          
          "                      <window:XXXXXXXX><start:XXXXXXXX> </length:XXXXXXXX>\n"
          "                      <port:XXX> </instrlen:XX> </fc:XX></skipdetect>\n"
          "                      </tinycfe></wiggler></verbose> \n\n"

          "            Required Parameter\n"
          "            ------------------\n"
          "            -backup:cfe\n"
          "            -backup:tfe\n"
          "            -backup:nvram\n"
          "            -backup:kernel\n"
          "            -backup:ternel\n"
          "            -backup:wholeflash\n"
          "            -backup:custom\n"
          "            -backup:bsp\n"
          "            -erase:cfe\n"
          "            -erase:tfe\n"
          "            -erase:nvram\n"
          "            -erase:kernel\n"
          "            -erase:ternel\n"
          "            -erase:wholeflash\n"
          "            -erase:custom\n"
          "            -erase:bsp\n"
          "            -flash:cfe\n"
          "            -flash:tfe\n"
          "            -flash:nvram\n"
          "            -flash:kernel\n"
          "            -flash:ternel\n"
          "            -flash:wholeflash\n"
          "            -flash:custom\n"
          "            -flash:bsp\n"
          "            -probeonly\n\n"

          "            Optional Switches\n"
          "            -----------------\n"
          "            /noreset ........... prevent Issuing EJTAG CPU reset\n"
          "            /noemw ............. prevent Enabling Memory Writes\n"
          "            /nocwd ............. prevent Clearing CPU Watchdog Timer\n"
          "            /nobreak ........... prevent Issuing Debug Mode JTAGBRK\n"
          "            /noerase ........... prevent Forced Erase before Flashing\n"
          "            /notimestamp ....... prevent Timestamping of Backups\n"
          "            /dma ............... force use of DMA routines\n"
          "            /nodma ............. force use of PRACC routines (No DMA)\n"
          "            /initcpu............ load CPU configuration code\n"
          "            /revendian.......... reverse chip endian\n"
          "            /window:XXXXXXXX ... custom flash window base&probe address(in HEX)\n"
          "            /start:XXXXXXXX .... custom start location (in HEX)\n"
          "            /length:XXXXXXXX ... custom length (in HEX)\n"
          "            /verbose............ scrolling display of data\n"
          "            /skipdetect ........ skip auto detection of CPU Chip ID\n"
          "            /instrlen:XX ....... set CPU instruction length manually\n"
          "            /wiggler ........... use wiggler cable\n"
          "            /bypass ............ unlock Spansion bypass mode & disable polling\n"
          "            /forcealign......... force erase address align with block boundary\n"         
          "            /erasechip.......... erase whole chip, only work with -probeonly \n"
          "            /clearppb........... erase Spansion PPB,only work with -probeonly\n"
          "            /showppb ........... show flash sector protection status\n"
          "                                 only work with -probeonly \n"
          "            /tinycfe............ one sector tiny CFE support(= TFE or TERNEL)\n"
          "            /port:XXX........... customize parallel port(default XXX is 378)\n"
          "                                 only work in Windows version\n\n"
          "            /fc:XX = Optional (Manual) Flash Chip Selection\n"
          "            -----------------------------------------------\n");

  while (flash_chip->vendid)
    {
      printf("            /fc:%02d ............. %-40.40s\n", ++counter, flash_chip->flash_part);
      flash_chip++;
    }

  printf( "\n\n");
  printf( " NOTES: 1) 'backup:', 'flash:' and 'erase:' - the source filename must exist as\n"
          "           follows: CFE.BIN, NVRAM.BIN, KERNEL.BIN, WHOLEFLASH.BIN or CUSTOM.BIN\n"
          "           BSP.BIN, TFE.BIN, TERNEL.BIN\n\n"

          "        2) If you have difficulty auto-detecting a particular flash part\n"
          "           you can manually specify your exact part using the /fc:XX option.\n\n"

          "        3) If you have difficulty with the older bcm47xx chips or when no CFE\n"
          "           is currently active/operational you may want to try both the\n"
          "           /noreset and /nobreak command line options together.  Some bcm47xx\n"
          "           chips *may* always require both these options to function properly.\n\n"

          "        4) When using this utility, usually it is best to type the command line\n"
          "           out, then plug in the router, and then hit <ENTER> quickly to avoid\n"
          "           the CPUs watchdog interfering with the EJTAG operations.\n\n"

          "        5) /bypass - enables Unlock bypass command for some AMD/Spansion type\n"
          "           flashes, it also disables polling\n\n"

          "        6) When enable tiny CFE option use 'CFE.BIN + /tinycfe' or TFE.BIN\n"
          "           filename. CFE and KERNEL address are calculated with one bottom\n"
          "           sector size. Almost all BCM63x8 box use one sector size CFE.\n\n"

          "        7) /initcpu allow load some code to initialize the CPU. This may help\n"
          "           BCM6358 open all EBI address access.\n\n"

          "        8) '-probeonly /window:xxxxxxxx /erasechip' allow choose a workable\n"
          "           sector address to erase whole chip. This may help on a bricked box\n"
          "           with bad CFE\n\n"
          
          "        9) /forcealign - enable erase sectors if the operation window is not\n"
          "           aligned with sector boundary. It's risky! but can help erase some\n"
          "           box NVRAM area whose sector size is large than NVRAM definition\n\n" 
          
          " ***************************************************************************\n"
          " * Flashing the KERNEL or WHOLEFLASH will take a very long time using JTAG *\n"
          " * via this utility.  You are better off flashing the CFE & NVRAM files    *\n"
          " * & then using the normal TFTP method to flash the KERNEL via ethernet.   *\n"
          " ***************************************************************************\n\n");
}


int main(int argc, char** argv)
{
  char choice[128];
  int run_option;
  int j;

  printf("\n");
  printf("===============================================\n");
  printf(" Broadcom EJTAG Debrick Utility v1.6r-hugebird\n");
  printf("===============================================\n\n");

  if (argc < 2)
    {
      show_usage();
      exit(1);
    }

  strcpy(choice,argv[1]);

  run_option = 0;

  if (strcasecmp(choice,"-backup:cfe")==0)
    {
      run_option = 1;
      strcpy(AREA_NAME, "CFE");
    }
  if (strcasecmp(choice,"-backup:tfe")==0)
    {
      run_option = 1;
      force_tinycfe = 1;
      strcpy(AREA_NAME, "TFE");
    }
  if (strcasecmp(choice,"-backup:nvram")==0)
    {
      run_option = 1;
      strcpy(AREA_NAME, "NVRAM");
    }
  if (strcasecmp(choice,"-backup:kernel")==0)
    {
      run_option = 1;
      strcpy(AREA_NAME, "KERNEL");
    }
  if (strcasecmp(choice,"-backup:ternel")==0)
    {
      run_option = 1;
      force_tinycfe = 1;
      strcpy(AREA_NAME, "TERNEL");
    }
  if (strcasecmp(choice,"-backup:wholeflash")==0)
    {
      run_option = 1;
      strcpy(AREA_NAME, "WHOLEFLASH");
    }
  if (strcasecmp(choice,"-backup:custom")==0)
    {
      run_option = 1;
      strcpy(AREA_NAME, "CUSTOM");
      custom_options++;
    }
  if (strcasecmp(choice,"-backup:bsp")==0)
    {
      run_option = 1;
      strcpy(AREA_NAME, "BSP");
    }

  if (strcasecmp(choice,"-erase:cfe")==0)
    {
      run_option = 2;
      strcpy(AREA_NAME, "CFE");
    }
  if (strcasecmp(choice,"-erase:tfe")==0)
    {
      run_option = 2;
      force_tinycfe = 1;
      strcpy(AREA_NAME, "TFE");
    }
  if (strcasecmp(choice,"-erase:nvram")==0)
    {
      run_option = 2;
      strcpy(AREA_NAME, "NVRAM");
    }
  if (strcasecmp(choice,"-erase:kernel")==0)
    {
      run_option = 2;
      strcpy(AREA_NAME, "KERNEL");
    }
  if (strcasecmp(choice,"-erase:ternel")==0)
    {
      run_option = 2;
      force_tinycfe = 1;
      strcpy(AREA_NAME, "TERNEL");
    }
  if (strcasecmp(choice,"-erase:wholeflash")==0)
    {
      run_option = 2;
      strcpy(AREA_NAME, "WHOLEFLASH");
    }
  if (strcasecmp(choice,"-erase:custom")==0)
    {
      run_option = 2;
      strcpy(AREA_NAME, "CUSTOM");
      custom_options++;
    }
  if (strcasecmp(choice,"-erase:bsp")==0)
    {
      run_option = 2;
      strcpy(AREA_NAME, "BSP");
    }

  if (strcasecmp(choice,"-flash:cfe")==0)
    {
      run_option = 3;
      strcpy(AREA_NAME, "CFE");
    }
  if (strcasecmp(choice,"-flash:tfe")==0)
    {
      run_option = 3;
      force_tinycfe = 1;
      strcpy(AREA_NAME, "TFE");
    }
  if (strcasecmp(choice,"-flash:nvram")==0)
    {
      run_option = 3;
      strcpy(AREA_NAME, "NVRAM");
    }
  if (strcasecmp(choice,"-flash:kernel")==0)
    {
      run_option = 3;
      strcpy(AREA_NAME, "KERNEL");
    }
  if (strcasecmp(choice,"-flash:ternel")==0)
    {
      run_option = 3;
      force_tinycfe = 1;
      strcpy(AREA_NAME, "TERNEL");
    }
  if (strcasecmp(choice,"-flash:wholeflash")==0)
    {
      run_option = 3;
      strcpy(AREA_NAME, "WHOLEFLASH");
    }
  if (strcasecmp(choice,"-flash:custom")==0)
    {
      run_option = 3;
      strcpy(AREA_NAME, "CUSTOM");
      custom_options++;
    }
  if (strcasecmp(choice,"-flash:bsp")==0)
    {
      run_option = 3;
      strcpy(AREA_NAME, "BSP");
    }

  if (strcasecmp(choice,"-probeonly")==0)
    {
      run_option = 4;
    }


  if (run_option == 0)
    {
      show_usage();
      printf("\n*** ERROR - Invalid [option] specified ***\n\n");
      exit(1);
    }

  if (argc > 2)
    {
      j = 2;
      while (j < argc)
        {
          strcpy(choice,argv[j]);
       /* **     CPU    ** */
          if (strcasecmp(choice,"/noreset")==0)              issue_reset = 0;
          else if (strcasecmp(choice,"/noemw")==0)           issue_enable_mw = 0;
          else if (strcasecmp(choice,"/nocwd")==0)           issue_watchdog = 0;
          else if (strcasecmp(choice,"/nobreak")==0)         issue_break = 0;
          else if (strcasecmp(choice,"/revendian")==0)       force_rev_endian = 1;
          else if (strcasecmp(choice,"/initcpu")==0)         issue_initcpu=1;
          else if (strcasecmp(choice,"/skipdetect")==0)      skipdetect = 1;
          else if (strncasecmp(choice,"/instrlen:",10)==0)   instrlen = strtoul(((char *)choice + 10),NULL,10);
 
       /* **    Flash   ** */
          else if (strcasecmp(choice,"/noerase")==0)         issue_erase = 0;
          else if (strcasecmp(choice,"/dma")==0)             force_dma = 1;
        else if (strcasecmp(choice,"/nodma")==0)           {force_nodma = 1; force_dma = 0; }
          else if (strcasecmp(choice,"/forcealign")==0) 	   force_align =1;
          else if (strncasecmp(choice,"/fc:",4)==0)          selected_fc = strtoul(((char *)choice + 4),NULL,10);
          else if (strcasecmp(choice,"/bypass")==0)          bypass = 1;
        else if (strcasecmp(choice,"/erasechip")==0)     { if (run_option ==4) issue_chip_erase = 1; } 
        else if (strcasecmp(choice,"/clearppb")==0)      { if (run_option ==4) issue_clear_ppb = 1;  }	
 				else if (strcasecmp(choice,"/showppb")==0)       { if (run_option ==4) issue_show_ppb = 1;   } 	
 						
       /* **    File   ** */
          else if (strcasecmp(choice,"/notimestamp")==0)     issue_timestamp = 0;	
          else if (strcasecmp(choice,"/tinycfe")==0)        
          	{  force_tinycfe = 1;
          		 if (strcasecmp(AREA_NAME,"CFE")==0)           strcpy(AREA_NAME, "TFE");
          		 else if (strcasecmp(AREA_NAME,"KERNEL")==0)   strcpy(AREA_NAME, "TERNEL");
    				}
          else if (strncasecmp(choice,"/window:",8)==0)
            {
              selected_window = strtoul(((char *)choice + 8),NULL,16);
              custom_options++;
              FLASH_PROBE_BASE = selected_window;
              if (run_option == 4) probeme = 1;
            }
          else if (strncasecmp(choice,"/start:",7)==0)
            {
              selected_start  = strtoul(((char *)choice + 7),NULL,16);
              custom_options++;
            }
          else if (strncasecmp(choice,"/length:",8)==0)
            {
              selected_length = strtoul(((char *)choice + 8),NULL,16);
              custom_options++;
            }
            
        /* **    Others   ** */    
          else if (strcasecmp(choice,"/verbose")==0)         silent_mode = 0;
          else if (strcasecmp(choice,"/wiggler")==0)         cable_type = WIGGLER;
#ifdef WINDOWS_VERSION   // ---- Compiler Specific Code ----
          else if (strncasecmp(choice,"/port:",6)==0)
            {
              lpt_port = strtoul(((char *)choice + 6),NULL,16);
            }
#endif
          else
            {
              show_usage();
              printf("\n*** ERROR - Input invalid <option> specified ***\n\n");
              exit(1);
            }
          j++;
        }
    }

  if (strcasecmp(AREA_NAME,"CUSTOM")==0)
    {
      if ((custom_options != 0) && (custom_options != 4))
        {
          show_usage();
          printf("\n*** ERROR - 'CUSTOM' also requires '/window' '/start' and '/length' options ***\n\n");
          exit(1);
        }
    }


  // ----------------------------------
  // Detect CPU
  // ----------------------------------
  chip_detect();


  // ----------------------------------
  // Find Implemented EJTAG Features
  // ----------------------------------
  check_ejtag_features();


  // ----------------------------------
  // Reset State Machine For Good Measure
  // ----------------------------------
  test_reset();


  // ----------------------------------
  // Reset Processor and Peripherals
  // ----------------------------------
  printf("Issuing Processor / Peripheral Reset ... ");
  if (issue_reset)
    {
      set_instr(INSTR_CONTROL);
      ctrl_reg = ReadWriteData(PRRST | PERRST);
      set_instr(INSTR_CONTROL);
      ctrl_reg = ReadWriteData(0x0);
      set_instr(INSTR_CONTROL);
      ctrl_reg = ReadWriteData(0x0);
      printf("Done\n");
    }
  else printf("Skipped\n");


  // ----------------------------------
  // Enable Memory Writes
  // ----------------------------------
  // Always skip for EJTAG versions 2.5 and 2.6 since they do not support DMA transactions.
  // Memory Protection bit only applies to EJTAG 2.0 based chips.
  if (ejtag_version != 0)  issue_enable_mw = 0;
  printf("Enabling Memory Writes ... ");
  if (issue_enable_mw)
    {
      // Clear Memory Protection Bit in DCR
      ejtag_dma_write(0xff300000, (ejtag_dma_read(0xff300000) & ~(1<<2)) );
      printf("Done\n");
    }
  else printf("Skipped\n");


  // ----------------------------------
  // Put into EJTAG Debug Mode
  // ----------------------------------
  printf("Halting Processor ... ");
  if (issue_break)
    {
      set_instr(INSTR_CONTROL);
      ctrl_reg = ReadWriteData(PRACC | PROBEN | SETDEV | JTAGBRK );
      if (ReadWriteData(PRACC | PROBEN | SETDEV) & BRKST)
        printf("<Processor Entered Debug Mode!> ... ");
      else
        printf("<Processor did NOT enter Debug Mode!> ... ");
      printf("Done\n");
    }
  else printf("Skipped\n");

  // ----------------------------------
  // Initialize data register
  // ----------------------------------
  
  if (!force_dma) sp_init_dreg();
      	
  // ----------------------------------
  // Clear Watchdog
  // ----------------------------------
  printf("Clearing Watchdog ... ");
  if (issue_watchdog)
    {
      ejtag_write(0xb8000080,0);
      printf("Done\n");
    }
  else printf("Skipped\n");

  // ----------------------------------
  // Initialize Processor Configuration
  // ----------------------------------
  printf("Loading CPU Configuration Code ... ");
  if (issue_initcpu) 
    {
      sp_init_cpu();
      printf("Done\n");
      if (!force_dma) sp_init_dreg();
    }
  else printf("Skipped\n");

  // ----------------------------------
  // Flash Chip Detection
  // ----------------------------------
  if (selected_fc != 0)
    sflash_config();
  else
    sflash_probe();


  // ----------------------------------
  // Execute Requested Operation
  // ----------------------------------
  if ((flash_size > 0) && (AREA_LENGTH > 0))
    {
      if (run_option == 1 )  run_backup(AREA_NAME, AREA_START, AREA_LENGTH);
      if (run_option == 2 )  run_erase(AREA_NAME, AREA_START, AREA_LENGTH);
      if (run_option == 3 )  run_flash(AREA_NAME, AREA_START, AREA_LENGTH);
    }
     
      if ((flash_size > 0) && (run_option == 4 ))
      	   { if (issue_chip_erase || issue_show_ppb || issue_clear_ppb) sp_check_ppb(); }
      	// In probe whole chip erase can be done at customed address.
      	// Check the Spansion advanced sector protect
    

  printf("\n\n *** REQUESTED OPERATION IS COMPLETE ***\n\n");

  chip_shutdown();

  return 0;
}


// **************************************************************************
// End of File
// **************************************************************************
