// *************************************************************************
// modified by hugebird @ chinadsl.net 12/12/2008 rev1.6
// *************************************************************************
// tornado@odessaua.com tornado@dd-wrt.com
// **************************************************************************
//  Written by HairyDairyMaid (a.k.a. - lightbulb)
//  hairydairymaid@yahoo.com
// **************************************************************************
//
//  This program is copyright (C) 2004 HairyDairyMaid (a.k.a. Lightbulb)
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of version 2 the GNU General Public License as published
//  by the Free Software Foundation.
//  This program is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
//  more details.
//  To view a copy of the license go to:
//  http://www.fsf.org/copyleft/gpl.html
//  To receive a copy of the GNU General Public License write the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// **************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#ifndef WINDOWS_VERSION

   #include <unistd.h>
   #include <sys/ioctl.h>

   #ifdef __FreeBSD__
      #include <dev/ppbus/ppi.h>
      #include <dev/ppbus/ppbconf.h>
      #define PPWDATA PPISDATA
      #define PPRSTATUS PPIGSTATUS
   #else
      #include <linux/ppdev.h>
   #endif

#endif

#define true  1
#define false 0


#define RETRY_ATTEMPTS 16
#define MAX_ATTEMPTS 50

/*
 * The followint table shows the pin assignment of 25-pin  Parallel Printer Port.
 * please refer to IEEE 1284 standard for detailed description.
 * data port (Out) (0x378)                 status port (In) (0x379)
 * bit[7] -- pin9 (Out)         bit[7] -- pin11 (In), busy (Hardware Inverted)
 * bit[6] -- pin8 (Out)         bit[6] -- pin10 (In), Ack
 * bit[5] -- pin7 (Out)         bit[5] -- pin12 (In), Paper out
 * bit[4] -- pin6 (Out)         bit[4] -- pin13 (In), Select
 * bit[3] -- pin5 (Out)         bit[3] -- pin15 (In), Error
 * bit[2] -- pin4 (Out)         bit[2] -- IRQ(Not)
 * bit[1] -- pin3 (Out)         bit[1] -- Reserved
 * bit[0] -- pin2 (Out)         bit[0] -- Reserved
 */
#define PORT378	0x378

// --- Xilinx Type Cable ---
#define TDI     0
#define TCK     1
#define TMS     2
#define TDO     4

// --- Wiggler Type Cable ---
#define WTDI      3
#define WTCK      2
#define WTMS      1
#define WTDO      7
#define WTRST_N   4

// --- Cable Type ---
#define XILINX    0
#define WIGGLER   1


// --- Some EJTAG Instruction Registers ---
#define INSTR_EXTEST    0x00
#define INSTR_IDCODE    0x01
#define INSTR_SAMPLE    0x02
#define INSTR_IMPCODE   0x03
#define INSTR_ADDRESS   0x08
#define INSTR_DATA      0x09
#define INSTR_CONTROL   0x0A
#define INSTR_BYPASS    0xFF

// --- Some EJTAG Bit Masks ---
#define TOF             (1 << 1 )
#define BRKST           (1 << 3 )      
#define DRWN            (1 << 9 )     
#define DERR            (1 << 10)
#define DSTRT           (1 << 11)
#define SETDEV          (1 << 14) //prob trap
#define PROBEN          (1 << 15) 
#define DMAACC          (1 << 17)
#define PRACC           (1 << 18) 
#define PRNW            (1 << 19) 
#define DLOCK           (1 << 5 )
#define TIF             (1 << 2 )
#define SYNC            (1 << 23)
#define PRRST           (1 << 16) 
#define PERRST          (1 << 20) 
#define JTAGBRK         (1 << 12) 
#define DNM             (1 << 28)
#define ROCC            (1 << 31) 

/* define data type for MIPS32 */
#define MIPS_BYTE        0 
#define MIPS_HALFWORD    1 
#define MIPS_WORD        2 
#define MIPS_TRIPLEBYTE  3 


#define DMA_BYTE        (MIPS_BYTE       << 7)  //DMA tranfser size one byte       0x00000000
#define DMA_HALFWORD    (MIPS_HALFWORD   << 7)  //DMA transfer size double bytes   0x00000080
#define DMA_WORD        (MIPS_WORD       << 7)  //DMA transfer size four bytes     0x00000100 
#define DMA_TRIPLEBYTE  (MIPS_TRIPLEBYTE << 7)  //DMA transfer size three bytes    0x00000180

#define END_MIPSLE			0		// Little Endian
#define END_MIPS	      1   // Big Endian

#define  size4K        0x1000
#define  size8K        0x2000
#define  size16K       0x4000
#define  size32K       0x8000
#define  size64K       0x10000
#define  size128K      0x20000
#define  size192K	     0x30000      // use for re-address tiny CFE and Kernel

#define  size1MB       0x100000
#define  size2MB       0x200000
#define  size4MB       0x400000
#define  size8MB       0x800000
#define  size16MB      0x1000000

#define  MEM_TOP			 0x20000000
#define  CFE_LEN       0x40000
#define  NVRAM_LEN     0x10000

#define  CMD_TYPE_BSC  0x01
#define  CMD_TYPE_SCS  0x02
#define  CMD_TYPE_AMD  0x03
#define  CMD_TYPE_SST  0x04

#define  SP_PPB        0x01      //Spansion Advanced Sector Protection

#define  STATUS_READY  0x0080    //DQ7


// EJTAG DEBUG Unit Vector on Debug Break
#define MIPS_DEBUG_VECTOR_ADDRESS           0xFF200200

// Our 'Pseudo' Virtual Memory Access Registers
#define MIPS_VIRTUAL_ADDRESS_ACCESS         0xFF200000        //not used by new pracc code
#define MIPS_VIRTUAL_DATA_ACCESS            0xFF200004


// --- Uhh, Just Because I Have To ---
void chip_detect(void);
void chip_shutdown(void);
static unsigned char clockin(int tms, int tdi);
void define_block(unsigned int block_count, unsigned int block_size);
static unsigned int ejtag_read(unsigned int addr);
static unsigned int ejtag_read_h(unsigned int addr);
void ejtag_write(unsigned int addr, unsigned int data);
void ejtag_write_h(unsigned int addr, unsigned int data);
static unsigned int ejtag_dma_read(unsigned int addr);
static unsigned int ejtag_dma_read_h(unsigned int addr);
void ejtag_dma_write(unsigned int addr, unsigned int data);
void ejtag_dma_write_h(unsigned int addr, unsigned int data);
static unsigned int ejtag_pracc_read(unsigned int addr);
void ejtag_pracc_write(unsigned int addr, unsigned int data);
static unsigned int ejtag_pracc_read_h(unsigned int addr);
void ejtag_pracc_write_h(unsigned int addr, unsigned int data);
static unsigned int ejtag_pracc_read2(unsigned int addr);
void ejtag_pracc_write2(unsigned int addr, unsigned int data);
static unsigned int ejtag_pracc_read_h2(unsigned int addr);
void ejtag_pracc_write_h2(unsigned int addr, unsigned int data);
static unsigned int *ejtag_fix_writecode( unsigned int addr, unsigned int data, int mode );
static unsigned int *ejtag_fix_readcode( unsigned int addr, int mode );
void identify_flash_part(void);
void lpt_closeport(void);
void lpt_openport(void);
static unsigned int ReadData(void);
static unsigned int ReadWriteData(unsigned int in_data);
void run_backup(char *filename, unsigned int start, unsigned int length);
void run_erase(char *filename, unsigned int start, unsigned int length);
void run_flash(char *filename, unsigned int start, unsigned int length);
void set_instr(int instr);
void sflash_config(void);
void sflash_erase_area(unsigned int start, unsigned int length);
void sflash_erase_block(unsigned int addr);
void sflash_probe(void);
void sflash_reset(void);
void sflash_poll(unsigned int addr, unsigned int data);
void sflash_write_word(unsigned int addr, unsigned int data);
void show_usage(void);
void ShowData(unsigned int value);
void ShowData_h(unsigned int value);
void test_reset(void);
void WriteData(unsigned int in_data);
void ExecuteDebugModule(unsigned int *pmodule);
void check_ejtag_features(void);
void sp_exit_cmdset(void);
void sp_exit_cmdset2(void);
void sp_get_chipid(void);
void sp_check_ppb(void);
void sp_reset(void);
void sp_init_cpu();
void sp_init_dreg();
void sp_unlock_bypass(void);
void sp_unlock_bypass_reset(void);
unsigned long rev_endian_h(unsigned long data);
unsigned long rev_endian(unsigned long data);

// *************** Original Hd's Pracc code ******************
unsigned int pracc_readword_code_module[] = {
               // #
               // # HairyDairyMaid's Assembler PrAcc Read Word Routine
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
  0x3C01FF20,  // lui $1,  0xFF20
  0x34210000,  // ori $1,  0x0000
               //
               // # Load R2 with the address for the read
  0x8C220000,  // lw $2,  ($1)
               //
               // # Load R3 with the word @R2
  0x8C430000,  // lw $3, 0($2)
               //
               // # Store the value into the pseudo-data register
  0xAC230004,  // sw $3, 4($1)
               //
  0x00000000,  // nop
  0x1000FFF9,  // beq $0, $0, start
  0x00000000}; // nop


unsigned int pracc_writeword_code_module[] = {
               // #
               // # HairyDairyMaid's Assembler PrAcc Write Word Routine
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
  0x3C01FF20,  // lui $1,  0xFF20
  0x34210000,  // ori $1,  0x0000
               //
               // # Load R2 with the address for the write
  0x8C220000,  // lw $2,  ($1)
               //
               // # Load R3 with the data from pseudo-data register
  0x8C230004,  // lw $3, 4($1)
               //
               // # Store the word at @R2 (the address)
  0xAC430000,  // sw $3,  ($2)
               //
  0x00000000,  // nop
  0x1000FFF9,  // beq $0, $0, start
  0x00000000}; // nop


unsigned int pracc_readhalf_code_module[] = {
               // #
               // # HairyDairyMaid's Assembler PrAcc Read HalfWord Routine
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
  0x3C01FF20,  // lui $1,  0xFF20
  0x34210000,  // ori $1,  0x0000
               //
               // # Load R2 with the address for the read
  0x8C220000,  // lw $2,  ($1)
               //
               // # Load R3 with the half word @R2
  0x94430000,  // lhu $3, 0($2)
               //
               // # Store the value into the pseudo-data register
  0xAC230004,  // sw $3, 4($1)
               //
  0x00000000,  // nop
  0x1000FFF9,  // beq $0, $0, start
  0x00000000}; // nop


unsigned int pracc_writehalf_code_module[] = {
               // #
               // # HairyDairyMaid's Assembler PrAcc Write HalfWord Routine
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
  0x3C01FF20,  // lui $1,  0xFF20
  0x34210000,  // ori $1,  0x0000
               //
               // # Load R2 with the address for the write
  0x8C220000,  // lw $2,  ($1)
               //
               // # Load R3 with the data from pseudo-data register
  0x8C230004,  // lw $3, 4($1)
               //
               // # Store the half word at @R2 (the address)
  0xA4430000,  // sh $3,  ($2)
               //
  0x00000000,  // nop
  0x1000FFF9,  // beq $0, $0, start
  0x00000000}; // nop

//   **************** hugebird new code ************************

unsigned int pracc_init_dreg[] = {
	             // #
               // # hugebird: PrAcc init data register
               // # 
	             // start:
	             // # keep $1 for all R/W operation
  0x3C01FF20,  // lui  $1,0xFF20  # li $1, MIPS_VIRTUAL_DATA_BASE
  0x00000000,
  0x1000FFFD,  // b  start
  0x00000000}; // nop
     
unsigned int pracc_readword_new[] = {
               // #
               // # hugebird: PrAcc Read Word Routine (4 sets)
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
 // 0x3C01FF20,// lui $1,  0xFF20
 // 0x34210000,// ori $1,  0x0000
               //
               // # Load R2 with the address %hi
  0x3C021F01,  // lui $2,  addr_hi   # p+0
               //
               // # Load R3 with the word @R2
  0x8C438100,  // lw $3, addr_lo($2) #p+1        <-   100011 00010 00011
               //
               // # Store the value into the pseudo-data register
  0xAC230004,  // sw $3, 4($1)       #p+2
               //
  0x00000000,  // nop   ,delay must have, wait data arrive.            
  0x1000FFFB,  // b  start
  0x00000000}; // nop

unsigned int pracc_readhalf_new[] = {
               // #
               // # hugebird: PrAcc Read Word Routine(4 sets)
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
 // 0x3C01FF20,// lui $1,  0xFF20
 // 0x34210000,// ori $1,  0x0000
               //
               // # Load R2 with the address %hi
  0x3C021F01,  // lui $2,  addr_hi    # p+0
               //
               // # Load R3 with the word @R2
  0x94438100,  // lhu $3, addr_lo($2) # p+1         <- 100101 00010 00011
               //
               // # Store the value into the pseudo-data register
  0xAC230004,  // sw $3, 4($1)        # p+2         
               //
  0x00000000,  // nop   ,delay must have, wait wait data arrive            
  0x1000FFFB,  // b  start
  0x00000000}; // nop

unsigned int pracc_writehalf_new[] = {
               // #
               // # hugebird : PrAcc Write Word Routine (5 sets)
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
  //0x3C01FF20,  // lui $1,  0xFF20
  //0x34210000,  // ori $1,  0x0000
               //
               // # Load R2 with the address %hi
  0x3C021F01,  // lui $2,  addr_hi    # p+0
               //
               // # Load R3 with the data
  0x34030010,  // li $3, data         # p+1
               //
               // # Store the half word at @R2 (the address)
  0xA4438100,  // sh $3,  addr_lo($2) # p+2    <- 101001 00010 00011
               //             
  0x1000FFFC,  // beq $0, $0, start
  0x00000000}; // nop

unsigned int pracc_writeword_new[] = {
               // #
               // # hugebird : PrAcc Write Word Routine (5 sets)
               // #
               // start:
               //
               // # Load R1 with the address of the pseudo-address register
  //0x3C01FF20,  // lui $1,  0xFF20
  //0x34210000,  // ori $1,  0x0000
               //
               // # Load R2 with the address %hi
  0x3C021F01,  // lui $2,  addr_hi        # p+0
               //
               // # Load R3 with the data
  0x3C030001,  // lui $3, data_hi         # p+1
  0x34631000,	 // ori	$3,$3,data_lo       # p+2      <-001101 00011 00011
               //
               // # Store the half word at @R2 (the address)
  0xAC438100,  // sw $3, addr_lo($2)      # p+3      
               //             
  0x1000FFFB,  // beq $0, $0, start
  0x00000000}; // nop
  
// **************************************************************************
//     hugeird : add cpu init configuration code below.
//               don't forget add point to CPU structure.
// **************************************************************************
unsigned int init_6358[] = {
               // #
               // # hugebird: initialize 6358 to fully access 16MB flash
               // #
               // start:
               //
  0x00000000,  // nop
  0x0000e021,  // move	gp,zero
  0x3c09ff40,  // lui	$9, 0xff40   
  0x3529000c,  // ori	$9, $9, 0x0c   
  0x4089b006,  // mtc0	$9, $22, 6  
               //
  0x00000000,  // nop
  0x1000FFF9,  // beq $0, $0, start
  0x00000000}; // nop


// **************************************************************************
// End of File
// **************************************************************************
